/*
  Copyright (C) 2000-2003 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: SkyHolidayCalculator.m,v 1.2 2003/07/11 16:42:43 jan Exp $

#include "SkyHolidayCalculator.h"
#include "common.h"

@implementation SkyHolidayCalculator

+ (SkyHolidayCalculator *)calculatorWithYear:(int)_year
                                    timeZone:(NSTimeZone *)_tz
                                userDefaults:(id)_ud
{
  return
    AUTORELEASE([[SkyHolidayCalculator alloc] initWithYear:_year
                                              timeZone:_tz
                                              userDefaults:_ud]);
}

- (id)init {
  if ((self = [super init])) {
    self->year = -1;
    self->holidays = nil;
    self->easter    = nil;
    self->firstMay  = nil;
  }
  return self;
}

- (id)initWithYear:(int)_year
          timeZone:(NSTimeZone *)_tz
      userDefaults:(id)_ud
{
  if ((self = [self init])) {
    self->year = _year;
    ASSIGN(self->timeZone,_tz);
    ASSIGN(self->userDefaults,_ud);
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->userDefaults);
  RELEASE(self->timeZone);
  RELEASE(self->holidays);
  RELEASE(self->easter);
  RELEASE(self->firstMay);
  RELEASE(self->firstAdvent);
  RELEASE(self->christmasEve);
  [super dealloc];
}
#endif

// accessors

- (void)setYear:(int)_year {
  if (self->year != _year) {
    RELEASE(self->holidays);     self->holidays    = nil;
    RELEASE(self->easter);       self->easter       = nil;
    RELEASE(self->firstMay);     self->firstMay     = nil;
    RELEASE(self->firstAdvent);  self->firstAdvent  = nil;
    RELEASE(self->christmasEve); self->christmasEve = nil;
  }
  self->year = _year;
}
- (int)year {
  return self->year;
}

- (void)setTimeZone:(NSTimeZone *)_tz {
  ASSIGN(self->timeZone,_tz);
}
- (NSTimeZone *)timeZone {
  return self->timeZone;
}

- (void)setUserDefaults:(NSUserDefaults *)_ud {
  // force reload
  [self setYear:self->year+1];
  [self setYear:self->year-1];
  ASSIGN(self->userDefaults,_ud);
}
- (NSUserDefaults *)userDefaults {
  return self->userDefaults;
}

- (void)setEaster:(NSCalendarDate *)_easter {
  ASSIGN(self->easter,_easter);
}

// calculation of easter
- (NSCalendarDate *)easter {
  // Ostern
  /* http://www.uni-bamberg.de/~ba1lw1/fkal.html#Algorithmus */
  if (self->easter == nil) {
    int y = self->year;
    unsigned m, n;
    int a, b, c, d, e;
    unsigned easterMonth, easterDay;

    if ((y > 1699) && (y < 1800)) {
      m = 23;
      n = 3;
    }
    else if ((y > 1799) && (y < 1900)) {
      m = 23;
      n = 4;
    }
    else if ((y > 1899) && (y < 2100)) {
      m = 24;
      n = 5;
    }
    else if ((y > 2099) && (y < 2200)) {
      m = 24;
      n = 6;
    }
    else
      return nil;
    
    a = y % 19;
    b = y % 4;
    c = y % 7;
    d = (19 * a + m) % 30;
    e = (2 * b + 4 * c + 6 * d + n) % 7;
  
    easterMonth = 3;
    easterDay   = 22 + d + e;
    if (easterDay > 31) {
      easterDay  -= 31;
      easterMonth = 4;
      if (easterDay == 26)
        easterDay = 19;
      if ((easterDay == 25) && (d == 28) && (a > 10))
        easterDay = 18;
    }

    [self setEaster:
          [NSCalendarDate dateWithYear:y month:easterMonth day:easterDay
                          hour:0 minute:0 second:0 timeZone:self->timeZone]];
  }
  return self->easter;
}

/* fixed hollidays */

- (void)setFirstMay:(NSCalendarDate *)_firstMay {
  ASSIGN(self->firstMay,_firstMay);
}
- (NSCalendarDate *)firstMay {
  // Erster Mai
  if (self->firstMay == nil) {
   [self setFirstMay:
         [NSCalendarDate dateWithYear:self->year month:5 day:1
                         hour:0 minute:0 second:0 timeZone:self->timeZone]];
  }
  return self->firstMay;
}

- (NSCalendarDate *)allSaintsDay {
  // Allerheiligen
  return [NSCalendarDate dateWithYear:self->year month:11 day:1
                         hour:0 minute:0 second:0 timeZone:self->timeZone];
}

- (void)setChristmasEve:(NSCalendarDate *)_day {
  ASSIGN(self->christmasEve,_day);
}
- (NSCalendarDate *)christmasEve {
  // Weihnachten
  if (self->christmasEve == nil)
    [self setChristmasEve:
          [NSCalendarDate dateWithYear:self->year month:12 day:24
                          hour:0 minute:0 second:0 timeZone:self->timeZone]];
  
  return self->christmasEve;
}

- (NSCalendarDate *)christmasDay {
  // Erster Weihnachtsfeiertag
  return [NSCalendarDate dateWithYear:self->year month:12 day:25
                         hour:0 minute:0 second:0 timeZone:self->timeZone];
}

- (NSCalendarDate *)boxingDay {
  // Zweiter Weihnachtsfeiertag
  return [NSCalendarDate dateWithYear:self->year month:12 day:26
                         hour:0 minute:0 second:0 timeZone:self->timeZone];
}

- (NSCalendarDate *)newYearsEve {
  // Silvester
  return [NSCalendarDate dateWithYear:self->year month:12 day:31
                         hour:0 minute:0 second:0 timeZone:self->timeZone];
}

- (NSCalendarDate *)newYear {
  // Neujahr
  return [NSCalendarDate dateWithYear:self->year month:1 day:1
                         hour:0 minute:0 second:0 timeZone:self->timeZone];
}

/* variable holidays */

- (NSCalendarDate *)goodFriday {
  // Karfreitag
  return [[self easter] dateByAddingYears:0 months:0 days:-2];
}
- (NSCalendarDate *)easterMonday {
  // Ostermontag
  return [[self easter] dateByAddingYears:0 months:0 days:1];
}

- (NSCalendarDate *)shrovetide {
  // Fastnacht
  return [[self easter] dateByAddingYears:0 months:0 days:-47];
}

- (NSCalendarDate *)ascension {
  // Christi Himmelfahrt
  return [[self easter] dateByAddingYears:0 months:0 days:39];
}

- (NSCalendarDate *)whitsun {
  // Pfingstsonntag
  return [[self easter] dateByAddingYears:0 months:0 days:49];
}
- (NSCalendarDate *)whitmonday {
  // Pfingstmontag
  return [[self easter] dateByAddingYears:0 months:0 days:50];
}

- (NSCalendarDate *)corpurChristi {
  // Fronleichnam
  return [[self easter] dateByAddingYears:0 months:0 days:60];
}

- (NSCalendarDate *)summertimeChange {
  // Sommerzeitumstellung
  NSCalendarDate* lastMarch =
    [NSCalendarDate dateWithYear:self->year month:3 day:31
                    hour:0 minute:0 second:0 timeZone:self->timeZone];
  return [lastMarch dateByAddingYears:0 months:0 days:-[lastMarch dayOfWeek]];
}

- (NSCalendarDate *)wintertimeChange {
  // Winterzeitumstellung
  int y    = self->year;
  int mDay = y < 1996 ? 30 : 31;
  NSCalendarDate* lastOct =
    [NSCalendarDate dateWithYear:self->year
                    month:(y < 1996 ? 9 : 10)
                    day:mDay
                    hour:0 minute:0 second:0 timeZone:self->timeZone];
  return [lastOct dateByAddingYears:0 months:0
                  days:-[lastOct dayOfWeek]];
}

- (NSCalendarDate *)mothersDay {
  // Muttertag
  return [[self firstMay] dateByAddingYears:0 months:0
                          days:(14 - [[self firstMay] dayOfWeek])];
}

- (void)setFirstAdvent:(NSCalendarDate *)_fadvent {
  ASSIGN(self->firstAdvent,_fadvent);
}
- (NSCalendarDate *)firstAdvent {
  // Erster Advent
  if (self->firstAdvent == nil)
    [self setFirstAdvent:
          [[self christmasEve] dateByAddingYears:0 months:0
                               days: (-21 - [[self christmasEve] dayOfWeek])]];

  return self->firstAdvent;
}

- (NSCalendarDate *)dayOfPrayerAndRepetance {
  // Buss- und Bettag
  return [[self firstAdvent] dateByAddingYears:0 months:0 days:-11];
}

- (NSCalendarDate *)deathsSunday {
  // Totensonntag
  return [[self firstAdvent] dateByAddingYears:0 months:0 days:-7];
}

- (NSCalendarDate *)secondAdvent {
  // zweiter Advent
  return [[self firstAdvent] dateByAddingYears:0 months:0 days:7];
}

- (NSCalendarDate *)thirdAdvent {
  // dritter Advent
  return [[self firstAdvent] dateByAddingYears:0 months:0 days:14];
}

- (NSCalendarDate *)fourthAdvent {
  // vierter Advent
  return [[self firstAdvent] dateByAddingYears:0 months:0 days:21];
}

// holidays dictionary

- (void)addHolidayKey:(NSString *)_dateKey defaultKey:(NSString *)_defKey
             withName:(NSString *)_key toList:(NSMutableDictionary *)_list
{
  NSString *flagKey;
    
  flagKey = @"scheduler_show_holiday_";
  flagKey = [flagKey stringByAppendingString:_defKey];

  if ([self->userDefaults boolForKey:flagKey]) {
    NSArray  *a;

    a = [_list objectForKey:_dateKey];
    if (a == nil)
      a = [NSArray array];
    if (![a containsObject:_key])
      a = [a arrayByAddingObject:_key];
    
    [_list setObject:a forKey:_dateKey];
  }
}

- (void)addHolidayKey:(NSString *)_dateKey withName:(NSString *)_key
               toList:(NSMutableDictionary *)_list {
  [self addHolidayKey:_dateKey defaultKey:_key withName:_key toList:_list];
}

- (void)addHoliday:(NSCalendarDate *)_day withName:(NSString *)_key
            toList:(NSMutableDictionary *)_list
{
  NSString *dateKey;
  
  dateKey = [_day descriptionWithCalendarFormat:@"%Y-%m-%d"];
  
  [self addHolidayKey:dateKey withName:_key toList:_list];
}

- (void)addHoliday:(NSCalendarDate *)_day withDefKey:(NSString *)_defKey
          withName:(NSString *)_key toList:(NSMutableDictionary *)_list
{
  NSString *dateKey;

  dateKey = [_day descriptionWithCalendarFormat:@"%Y-%m-%d"];

  [self addHolidayKey:dateKey defaultKey:_defKey withName:_key toList:_list];
}

- (void)addEveryYearHolidays:(NSDictionary *)_days
                  withDefKey:(NSString *)_defKey
                      toList:(NSMutableDictionary *)_list
{
  NSArray *keys;
  int     i, cnt;

  keys = [_days allKeys];
      
  for (i = 0, cnt = [keys count]; i < cnt; i++) {
    NSString *key;
    NSString *dateKey;

    key = [keys objectAtIndex:i];
    dateKey = [NSString stringWithFormat:@"%i-%@", self->year, key];
    
    if (_defKey == nil) {
      [self addHolidayKey:dateKey withName:[_days objectForKey:key]
            toList:_list];
    }
    else {
      [self addHolidayKey:dateKey defaultKey:_defKey
            withName:[_days objectForKey:key]
            toList:_list];
    }
  }
}

- (void)setHolidays:(NSDictionary *)_holidays {
  ASSIGN(self->holidays,_holidays);
}
- (NSDictionary *)holidays {
  if (self->holidays == nil) {
    NSMutableDictionary *md = [NSMutableDictionary dictionaryWithCapacity:16];
    NSDictionary        *everyYear;
    NSDictionary        *thisYear;
    NSDictionary        *source = nil;
    NSDictionary        *customEveryYear;
    NSDictionary        *custom;
    NSString* file;

    file =
      [[[NSProcessInfo processInfo] environment]
                       objectForKey:@"GNUSTEP_LOCAL_ROOT"];
    file = [file stringByAppendingPathComponent:@"Libraries"];
    file = [file stringByAppendingPathComponent:@"Resources"];
    file = [file stringByAppendingPathComponent:@"OGoScheduler"];
    file = [file stringByAppendingPathComponent:@"Holidays.plist"];
    if (file != nil)
      source = [NSDictionary dictionaryWithContentsOfFile:file];

    everyYear =
      [source objectForKey:@"scheduler_everyyear_holidays"];
    thisYear =
      [source objectForKey:
            [NSString stringWithFormat:@"scheduler_%i_holidays", self->year]];

    customEveryYear = [self->userDefaults dictionaryForKey:
                           @"scheduler_custom_everyyear_holidays"];
    custom          = [self->userDefaults dictionaryForKey:
                           @"scheduler_custom_holidays"];

    // adding custom everyYear holidays
    if (customEveryYear != nil) {
      [self addEveryYearHolidays:customEveryYear
            withDefKey:@"holidaygroup_custom_private" toList:md];
    }
    // adding custom holidays
    if (custom != nil) {
      NSDictionary *thisYear =
        [custom objectForKey:[NSString stringWithFormat:@"%i",self->year]];
      if (thisYear != nil) {
        [self addEveryYearHolidays:thisYear
              withDefKey:@"holidaygroup_custom_private" toList:md];
      }
    }

    // adding everyYear holidays
    if (everyYear != nil) {
      NSArray *allGrps;
      int i, cnt;

      allGrps = [everyYear allKeys];

      for (i = 0, cnt = [allGrps count]; i < cnt; i++) {
        id key, val;
        NSString *defKey = nil;
        
        key = [allGrps objectAtIndex:i];
        val = [everyYear objectForKey:key];
        if ([val isKindOfClass:[NSString class]]) {
          val = [NSDictionary dictionaryWithObject:val forKey:key];
        }
        else {
          defKey = [NSString stringWithFormat:@"holidaygroup_%@", key];
        }
        [self addEveryYearHolidays:val withDefKey:defKey toList:md];
      }
    }

    // adding thisYear holidays
    if (thisYear != nil) {
      NSArray *keys;
      int      i, cnt;
      
      keys = [thisYear allKeys];
      
      for (i = 0, cnt = [keys count]; i < cnt; i++) {
        NSArray      *holiKeys;
        NSDictionary *holiDict;
        NSString     *key;
        int          j, holiCnt;

        key      = [keys objectAtIndex:i];
        holiDict = [thisYear objectForKey:key];
        holiKeys = [holiDict allKeys];

        for (j = 0, holiCnt = [holiKeys count]; j < holiCnt; j++) {
          NSString *dateKey;
          NSString *holiKey;
          NSString *holiName;
          NSString *defaultKey;

          holiKey  = [holiKeys objectAtIndex:j];
          dateKey  = [NSString stringWithFormat:@"%i-%@", self->year, holiKey];
          // custom all is non localized
          holiName = ([key isEqualToString:@"custom_all"])
            ? [holiDict objectForKey:holiKey]
            : [NSString stringWithFormat:@"%@_%@",
                        [holiDict objectForKey:holiKey], key];
          defaultKey = @"holidaygroup_";
          defaultKey = [defaultKey stringByAppendingString:key];

          [self addHolidayKey:dateKey defaultKey:defaultKey
                withName:holiName toList:md];
        }
      }
    }

    // Ostern
    [self addHoliday:[self goodFriday] withDefKey:@"holidaygroup_bylaw"
          withName:@"goodFriday" toList:md];
    [self addHoliday:[self easter] withDefKey:@"holidaygroup_bylaw"
          withName:@"easter" toList:md];
    [self addHoliday:[self easterMonday] withDefKey:@"holidaygroup_bylaw"
          withName:@"easterMonday" toList:md];
    // Fastnacht
    [self addHoliday:[self shrovetide] withName:@"shrovetide" toList:md];
    // Christi Himmelfahrt
    [self addHoliday:[self ascension] withDefKey:@"holidaygroup_bylaw"
          withName:@"ascension" toList:md];
    // Pfingsten
    [self addHoliday:[self whitsun] withDefKey:@"holidaygroup_bylaw"
          withName:@"whitsun" toList:md];
    [self addHoliday:[self whitmonday] withDefKey:@"holidaygroup_bylaw"
          withName:@"whitmonday" toList:md];
    // Erster Mai
    [self addHoliday:[self firstMay] withDefKey:@"holidaygroup_bylaw"
          withName:@"firstMay" toList:md];
    // Muttertag
    [self addHoliday:[self mothersDay] withName:@"mothersDay" toList:md];
    // Fronleichnam
    [self addHoliday:[self corpurChristi] withDefKey:@"holidaygroup_bylaw"
          withName:@"corpusChristi" toList:md];
    // Sommerzeitumstellung
    [self addHoliday:[self summertimeChange]
          withName:@"summertimeChange" toList:md];
    // Winterzeitumstellung
    [self addHoliday:[self wintertimeChange]
          withName:@"wintertimeChange" toList:md];
    // Weihnachtsabend
    [self addHoliday:[self christmasEve]
          withName:@"christmasEve"     toList:md];
    // Buss- und Bettag
    [self addHoliday:[self dayOfPrayerAndRepetance]
          withDefKey:@"holidaygroup_bylaw" withName:@"dayOfPrayerAndRepetance"
          toList:md];
    // Totensonntag
    [self addHoliday:[self deathsSunday]
          withName:@"deathsSunday"     toList:md];
    // Erster Advent
    [self addHoliday:[self firstAdvent] withDefKey:@"holidaygroup_advent"
          withName:@"firstAdvent" toList:md];
    // Zweiter Advent
    [self addHoliday:[self secondAdvent] withDefKey:@"holidaygroup_advent"
          withName:@"secondAdvent" toList:md];
    // Dritter Advent
    [self addHoliday:[self thirdAdvent] withDefKey:@"holidaygroup_advent"
          withName:@"thirdAdvent" toList:md];
    // Vierter Advent
    [self addHoliday:[self fourthAdvent] withDefKey:@"holidaygroup_advent"
          withName:@"fourthAdvent" toList:md];

    [self setHolidays:md];
  }
  return self->holidays;
}

- (NSArray *)holidaysOfDate:(NSCalendarDate *)_date {
  NSString *dateKey;

  if ((dateKey = [_date descriptionWithCalendarFormat:@"%Y-%m-%d"]) == nil)
    return nil;
  
  [self setYear:[_date yearOfCommonEra]];
  return [[self holidays] objectForKey:dateKey];
}

@end
