/*
  Copyright (C) 2000-2003 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: LSMoveJobCommand.m,v 1.2 2003/07/11 16:35:52 jan Exp $

#import <LSFoundation/LSDBObjectSetCommand.h>

@class NSNumber;

@interface LSMoveJobCommand : LSDBObjectSetCommand
{
@private
  id       sourceParent;
  id       destinationParent;
  id       destination;
  NSNumber *insertBehind;
  NSNumber *moveToTop;
  NSNumber *moveUp;
}

@end

#import "common.h"

@interface LSMoveJobCommand(PrivateMethodes)
- (void)_deleteOldJobAssignmentInContext:(id)_context;
- (void)_createNewJobAssignmentInContext:(id)_context;
- (void)_moveUpOrDownInContext:(id)_context;

- (id)_getParentOf:(id)_job inContext:(id)_context;
- (void)_insureChildPositionsOn:(id)_object inContext:(id)_context;
- (NSArray *)_getChildAssignsOf:(id)_object inContext:(id)_context;
@end

@implementation LSMoveJobCommand

static int comparePos(id jobAss1, id jobAss2, void *context) {
  NSNumber *pos1 = [jobAss1 valueForKey:@"position"];
  NSNumber *pos2 = [jobAss2 valueForKey:@"position"];
  
  if (pos1 == nil)
    return NSOrderedAscending;
  if (pos2 == nil)
    return NSOrderedDescending;
  return [pos1 compare:pos2];
}

- (id)initForOperation:(NSString *)_operation inDomain:(NSString *)_domain {
  if ((self = [super initForOperation:_operation inDomain:_domain])) {
    self->insertBehind = [NSNumber numberWithBool:YES];
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->sourceParent);
  RELEASE(self->destinationParent);
  RELEASE(self->destination);
  RELEASE(self->insertBehind);
  RELEASE(self->moveToTop);
  RELEASE(self->moveUp);
  
  [super dealloc];
}
#endif

// accessors

- (void)setInsertBehind:(NSNumber *)_insertBehind {
  ASSIGN(self->insertBehind, _insertBehind);
}
- (NSNumber *)insertBehind {
  return self->insertBehind;
}

- (void)setDestination:(id)_destination {
  ASSIGN(self->destination, _destination);
}
- (id)destination {
  return self->destination;
}

- (void)setMoveToTop:(NSNumber *)_moveToTop {
  ASSIGN(self->moveToTop, _moveToTop);
}
- (NSNumber *)moveToTop {
  return self->moveToTop;
}

- (void)setMoveUp:(NSNumber *)_moveUp {
  ASSIGN(self->moveUp, _moveUp);
}
- (id)moveUp {
  return self->moveUp;
}

// -----------------------------------------------------------------------

- (void)_validateKeysForContext:(id)_context {

  [self assert:(![[self destination] isEqual:[self object]])
        reason:@"source and destination are equal"];
  
  if ((self->moveUp == nil) && (self->moveToTop == nil)) {
    [self assert:[[self destination] isNotNull]
          reason:@"destination not set!"];
  }
  
  [super _validateKeysForContext:_context];
}

- (void)_prepareForExecutionInContext:(id)_context {
  id src = [self _getParentOf:[self object] inContext:_context];

  [self _insureChildPositionsOn:src inContext:_context];
  ASSIGN(self->sourceParent, src);

  if (self->destination != nil) {
    id dest = [self _getParentOf:[self destination] inContext:_context];

    [self _insureChildPositionsOn:dest inContext:_context];    
    ASSIGN(self->destinationParent, dest);
  }
}

- (void)_executeInContext:(id)_context {
  [super _executeInContext:_context];

  if ((self->moveToTop != nil) || (self->moveUp != nil)) {
    [self _moveUpOrDownInContext:_context];
  }
  else {
    [self _deleteOldJobAssignmentInContext:_context];
    [self _createNewJobAssignmentInContext:_context];
  }
  LSRunCommandV(_context, @"job", @"get",
                @"jobId", [[self object] valueForKey:@"jobId"], nil);
}

// initialize records

- (NSString *)entityName {
  return @"Job";
}

// key/value coding

- (void)takeValue:(id)_value forKey:(id)_key {
  if ([_key isEqualToString:@"destination"]) {
    [self setDestination:_value];
    return;
  }
  else if ([_key isEqualToString:@"moveToTop"]) {
    [self setMoveToTop:_value];
    return;
  }
  else if ([_key isEqualToString:@"moveToBottom"]) {
    [self setMoveToTop:[NSNumber numberWithBool:![_value boolValue]]];
    return;
  }
  else if ([_key isEqualToString:@"moveUp"]) {
    [self setMoveUp:_value];
    return;
  }
  else if ([_key isEqualToString:@"moveDown"]) {
    [self setMoveUp:[NSNumber numberWithBool:![_value boolValue]]];
    return;
  }  
  else if ([_key isEqualToString:@"insertBehind"]) {
    [self setInsertBehind:_value];
    return;
  }
  [super takeValue:_value forKey:_key];
}

- (id)valueForKey:(id)_key {
  if ([_key isEqualToString:@"destination"])
    return [self destination];
  else if ([_key isEqualToString:@"insertBehind"])
    return [self insertBehind];
  return [super valueForKey:_key];
}

// --- LSMoveJobCommand(PrivateMethodes) -----------------------------------

- (NSArray *)_getChildAssignsOf:(id)_object inContext:(id)_context {
  return LSRunCommandV(_context, @"job", @"get-subprocesses",
                       @"object", _object, nil);
}

- (id)_getParentOf:(id)_job inContext:(id)_context {
  NSArray *pList = LSRunCommandV(_context, @"job", @"get-parentprocess",
                                 @"object", _job, nil);
  return [[pList lastObject] valueForKey:@"toParentJob"];
}

- (void)_insureChildPositionsOn:(id)_object inContext:(id)_context {
  id  assigns = [self _getChildAssignsOf:_object inContext:_context];
  int i, cnt  = [assigns count];

  assigns = [assigns sortedArrayUsingFunction:comparePos context:NULL];
  
  for (i=0; i<cnt; i++) {
    id  assign = [assigns objectAtIndex:i];
    int pos    = [[assign valueForKey:@"position"] intValue];

    if (i+1 != pos) {
      [assign takeValue:[NSNumber numberWithInt:i+1] forKey:@"position"];
      LSRunCommandV(_context, @"jobassignment", @"set", @"object", assign, nil);
    }
  }
}

- (void)_moveUpOrDownInContext:(id)_context {
  id       obj     = [self object];
  NSArray *assigns = [self _getChildAssignsOf:sourceParent inContext:_context];
  id       src     = [[obj valueForKey:@"toParentJobAssignment"] lastObject];
  id       dest    = nil;
  int      srcPos  = [[src valueForKey:@"position"] intValue];
  int      destPos;

  assigns = [assigns sortedArrayUsingFunction:comparePos context:NULL];
  
  if (self->moveToTop != nil)
    destPos = ([self->moveToTop boolValue]) ? 1 : [assigns count];
  else
    destPos = ([self->moveUp boolValue]) ? srcPos-1 : srcPos+1;

  if (srcPos != destPos && 0 < destPos  && destPos <= [assigns count]) {
    dest = [assigns objectAtIndex:destPos-1];
    
    [src  takeValue:[NSNumber numberWithInt:destPos] forKey:@"position"];
    [dest takeValue:[NSNumber numberWithInt:srcPos]  forKey:@"position"];
    
    LSRunCommandV(_context, @"jobassignment", @"set", @"object", src, nil);
    LSRunCommandV(_context, @"jobassignment", @"set", @"object", dest, nil);
  }
}

@end
