/*
  Copyright (C) 2000-2004 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: NGImap4ResponseNormalizer.m,v 1.1 2004/01/24 23:46:41 helge Exp $

#include "NGImap4ResponseNormalizer.h"
#include "NGImap4Client.h"
#include "common.h"

@interface NGImap4Client(UsedPrivates)
- (NSString *)delimiter;
- (NSString *)_imapFolder2Folder:(NSString *)_folder;
@end

@implementation NGImap4ResponseNormalizer

static __inline__ NSArray *
_imapFlags2Flags(NGImap4ResponseNormalizer *, NSArray *);

static NSDictionary *VersionPrefixDict = nil;

static NSNumber *YesNumber     = nil;
static NSNumber *NoNumber      = nil;
static int      LogImapEnabled = -1;

+ (void)initialize {
  NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
  static BOOL didInit = NO;
  if (didInit) return;
  didInit = YES;
  
  YesNumber = [[NSNumber numberWithBool:YES] retain];
  NoNumber  = [[NSNumber numberWithBool:NO]  retain];

  LogImapEnabled     = [ud boolForKey:@"ImapLogEnabled"]?1:0;

  /*
    cyrus   - * OK defiant Cyrus IMAP4 v2.0.16 server ready
    courier - * OK Courier-IMAP ready. Copyright 1998-2002 Double
                Precision, Inc.  See COPYING for distribution information.
  */
  if (VersionPrefixDict == nil) {
    VersionPrefixDict =
      [[NSDictionary alloc] initWithObjectsAndKeys:
			      @"cyrus imap4 v", @"cyrus",
                              @" imap4rev1 ",   @"washington", 
                              @"courier",       @"courier", nil];
  }
}

- (id)initWithClient:(NGImap4Client *)_client {
  if ((self = [super init])) {
    self->client = _client; /* non-retained */
  }
  return self;
}

/* client callbacks */

- (void)closeConnection {
  [(id)self->client closeConnection];
}

- (NSString *)delimiter {
  return [self->client delimiter];
}

/* folder handling */

- (NSString *)_imapFolder2Folder:(NSString *)_folder {
  return [self->client _imapFolder2Folder:_folder];
}

/* primary */

- (NSMutableDictionary *)normalizeResponse:(NGHashMap *)_map {
  /*
    Filter for all responses
    result  : NSNumber (response result)
    exists  : NSNumber (number of exists mails in selected folder
    recent  : NSNumber (number of recent mails in selected folder
    expunge : NSArray  (message sequence number of expunged mails in selected 
                        folder)
  */
  NSMutableDictionary *result;
  id                  obj;
  NSDictionary        *respRes;
  
  if (_map == nil)
    return (id)[NSMutableDictionary dictionary];
  
  respRes = [[_map objectEnumeratorForKey:@"ResponseResult"] nextObject];
  result  = [NSMutableDictionary dictionaryWithCapacity:32];
  [result setObject:_map forKey:@"RawResponse"];
  
  if ((obj = [_map objectForKey:@"bye"])) {
    [result setObject:NoNumber forKey:@"result"];
    [result setObject:obj forKey:@"reason"];
    [self closeConnection];
    return result;
  }

  if ([[respRes objectForKey:@"result"] isEqual:@"ok"]) {
    [result setObject:YesNumber forKey:@"result"];
  }
  else {
    id tmp = nil;
    [result setObject:NoNumber forKey:@"result"];
    if ((tmp = [respRes objectForKey:@"description"]) != nil) {
      [result setObject:tmp forKey:@"reason"];
    }
    return result;
  }
  if ((obj = [[_map objectEnumeratorForKey:@"exists"] nextObject]) != nil) { // 
    [result setObject:obj forKey:@"exists"];
  }
  if ((obj = [[_map objectEnumeratorForKey:@"recent"] nextObject]) != nil) {
    [result setObject:obj forKey:@"recent"];
  }
  if ((obj = [_map objectsForKey:@"expunge"]) != nil)
    [result setObject:obj forKey:@"expunge"];
  
  return result;
}

- (NSDictionary *)normalizeSortResponse:(NGHashMap *)_map {
  /* filter for sort response (search  : NSArray (msn)) */
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"sort"] nextObject]))
    [result setObject:obj forKey:@"sort"];
  
  return result;
}

- (NSDictionary *)normalizeCapabilityRespone:(NGHashMap *)_map {
  /* filter for capability response: capability  : NSArray */
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"capability"] nextObject]))
    [result setObject:obj forKey:@"capability"];
  
  return result;
}

- (NSDictionary *)normalizeThreadResponse:(NGHashMap *)_map {
  /* filter for thread response: thread  : NSArray (msn) */
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"thread"] nextObject]))
    [result setObject:obj forKey:@"thread"];
  
  return result;
}

- (NSDictionary *)normalizeSearchResponse:(NGHashMap *)_map {
  /* filter for search response: search  : NSArray (msn) */
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"search"] nextObject]))
    [result setObject:obj forKey:@"search"];
  
  return result;
}

- (NSDictionary *)normalizeSelectResponse:(NGHashMap *)_map {
  /*
    filter for select response
      flags  : NSArray
      unseen : NSNumber
      access  : NSString ([READ-WRITE], ... )
  */
  id                  obj;
  NSEnumerator        *enumerator;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"flags"] nextObject]))
    [result setObject:_imapFlags2Flags(self, obj) forKey:@"flags"];
  
  enumerator = [_map objectEnumeratorForKey:@"ok"];
  while ((obj = [enumerator nextObject])) {
    id o;
    
    if ((o = [obj  objectForKey:@"unseen"]))
      [result setObject:o forKey:@"unseen"];
  }
  
  enumerator = [_map objectEnumeratorForKey:@"no"];
  while ((obj = [enumerator nextObject])) {
    id o;
    if ([obj isKindOfClass:[NSDictionary class]]) {
      if ((o = [obj  objectForKey:@"ALERT"]))
        [result setObject:o forKey:@"alert"];
    }
    else {
      [result setObject:obj forKey:@"alert"];
    }
  }
  
  if ((obj = [[_map objectForKey:@"ResponseResult"] objectForKey:@"flag"]))
    [result setObject:obj forKey:@"access"];
  
  return result;
}

- (NSDictionary *)normalizeStatusResponse:(NGHashMap *)_map {
  /*
    filter for status response
      messages  : NSNumber
      recent    : NSNumber
      unseen    : NSNumber
  */
  id                  obj;
  NSMutableDictionary *result;
  id                  o;
  
  result = [self normalizeResponse:_map];
  
  obj = [[[_map objectEnumeratorForKey:@"status"] nextObject]
                objectForKey:@"flags"];

  if ((o = [obj  objectForKey:@"messages"]) != nil) {
    [result setObject:o forKey:@"messages"];
  }
  if ((o = [obj  objectForKey:@"recent"]) != nil) {
    if ([result objectForKey:@"recent"] == nil)
      [result setObject:o forKey:@"recent"];
  }
  if ((o = [obj  objectForKey:@"unseen"]) != nil) {
    [result setObject:o forKey:@"unseen"];
  }
  return result;
}

/*
** filter for fetch response
**       fetch : NSArray (fetch responses)
**             'header'  - RFC822.HEADER
**             'text'    - RFC822.TEXT
**             'size'    - SIZE
**             'flags'   - FLAGS
**             'uid'     - UID
**             'msn'     - message sequence number
**             'message' - RFC822
**             'body   ' - (dictionary with bodystructure)
*/

- (NSDictionary *)normalizeFetchResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;
  NSEnumerator        *enumerator;
  NSMutableArray      *fetch;
  
  result = [self normalizeResponse:_map];
  fetch  = [NSMutableArray arrayWithCapacity:512];
  
  enumerator = [_map objectEnumeratorForKey:@"fetch"];
  
  while ((obj = [enumerator nextObject])) {
    NSDictionary *entry;
    NSEnumerator *keyEnum;
    id           key;
    id           keys[9];
    id           values[9];
    unsigned     count;
    
    keyEnum = [obj keyEnumerator];    
    count   = 0;
    
    while ((key = [keyEnum nextObject]) && (count < 9)) {
      if ([key isEqualToString:@"rfc822"]) {
        keys[count]   = @"message";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.header"]) {
        keys[count]   = @"header";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.text"]) {
        keys[count]   = @"text";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.size"]) {
        keys[count]   = @"size";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"flags"]) {
        keys[count]   = @"flags";
        values[count] = _imapFlags2Flags(self, [obj objectForKey:key]);
        count++;
      }
      else if ([key isEqualToString:@"uid"]) {
        keys[count]   = @"uid";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"msn"]) {
        keys[count]   = @"msn";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([(NSString *)key hasPrefix:@"body"]) {
        keys[count]   = @"body";
        values[count] = [obj objectForKey:key];
        count++;
      }
    }
    
    /* create dictionary */
    
    entry =
      [[NSDictionary alloc] initWithObjects:values forKeys:keys count:count];
    
    if (entry) {
      [fetch addObject:entry];
      [entry release]; entry = nil;
    }
  }
  [result setObject:fetch forKey:@"fetch"];
  return [[result copy] autorelease];
}

/*
** filter for quota responses
*/

- (NSDictionary *)normalizeQuotaResponse:(NGHashMap *)_map {
  NSMutableDictionary *result, *quotaRoot, *quota, *tmp;
  id                  obj;
  NSEnumerator        *enumerator;

  result     = [self normalizeResponse:_map];
  quotaRoot  = [_map objectForKey:@"quotaRoot"];
  quota      = [_map objectForKey:@"quota"];
  enumerator = [quotaRoot keyEnumerator];
  tmp        = [NSMutableDictionary dictionaryWithCapacity:[quota count]];
  
  while ((obj = [enumerator nextObject])) {
    NSString     *qRoot;
    NSDictionary *qDesc;

    qRoot = [quotaRoot objectForKey:obj];

    if (![qRoot length]) {
      if (LogImapEnabled) {
        [self logWithFormat:@"%s: missing quotaroot for %@",
              __PRETTY_FUNCTION__, obj];
      }
      continue;
    }
    qDesc = [quota objectForKey:qRoot];

    if (![qDesc count]) {
      if (LogImapEnabled) {
        [self logWithFormat:@"%s: missing quota description for"
              @" folder %@ root %@",
              __PRETTY_FUNCTION__, obj, qRoot];
      }
      continue;
    }
    [tmp setObject:qDesc forKey:[self _imapFolder2Folder:obj]];
  }
  [result setObject:tmp forKey:@"quotas"];
  return [[result copy] autorelease];
}


/*
** filter for open connection
*/

- (NSDictionary *)normalizeOpenConnectionResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;

  result = [self normalizeResponse:_map];
  
  obj = [[_map objectEnumeratorForKey:@"ok"] nextObject];
  if (obj == nil) {
    [result setObject:NoNumber forKey:@"result"];
    return result;
  }
  
  if ([obj isKindOfClass:[NSDictionary class]])
    obj = [obj objectForKey:@"comment"];

  if ([obj isKindOfClass:[NSString class]]) {
    NSEnumerator *enumerator;
    id           key;
    NSString     *lowServer;
      
    [result setObject:obj forKey:@"server"];
      
    enumerator = [VersionPrefixDict keyEnumerator];
    lowServer  = [obj lowercaseString];

    while ((key = [enumerator nextObject])) {
      NSString *pref;
      NSArray  *vers;
      NSRange  r;
        
      pref  = [VersionPrefixDict objectForKey:key];
      r     = [lowServer rangeOfString:pref];
      
      if (r.length == 0) continue;
        
      [result setObject:key forKey:@"serverKind"];
      if (![key isEqualToString:@"cyrus"])
        continue;
        
      /* cyrus server, collect version */
        
      vers = [[lowServer substringFromIndex:(r.location + [pref length])]
                         componentsSeparatedByString:@"."];
      
      if ([vers count] > 2) {
        NSNumber *n;
          
        n = [NSNumber numberWithInt:[[vers objectAtIndex:0] intValue]];
        [result setObject:n forKey:@"version"];

        n = [NSNumber numberWithInt:[[vers objectAtIndex:1] intValue]];
        [result setObject:n forKey:@"subversion"];
          
        n = [NSNumber numberWithInt:[[vers objectAtIndex:2] intValue]];
        [result setObject:n forKey:@"tag"];
      }
      break;
    }
  }
  [result setObject:YesNumber forKey:@"result"];
  
  return result;
}

/*
** filter for list
**       list : NSDictionary (folder name as key and flags as value)
*/

- (NSDictionary *)normalizeListResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;
  NSAutoreleasePool   *pool;
  NSDictionary        *rr;
  
  pool   = [[NSAutoreleasePool alloc] init];
  result = [self normalizeResponse:_map];
  
  if ((obj = [_map objectsForKey:@"list"]) != nil) {
    NSEnumerator        *enumerator = nil;
    id                  o           = nil;
    NSMutableDictionary *folder     = nil;    
    
    enumerator = [obj objectEnumerator];
    folder     = [[NSMutableDictionary alloc] init];
    
    while ((o = [enumerator nextObject])) {
      [folder setObject:_imapFlags2Flags(self, [o objectForKey:@"flags"])
              forKey:[self _imapFolder2Folder:[o objectForKey:@"folderName"]]];
    }
    
    {
      NSDictionary *f;
      
      f = [folder copy];
      [result setObject:f forKey:@"list"];
      [f release];      f      = nil;
      [folder release]; folder = nil;
    }
  }
  rr = [result copy];
  [pool release];
  
  return [rr autorelease];
}

/* flags */

static inline NSArray *
_imapFlags2Flags(NGImap4ResponseNormalizer *self, NSArray *_flags) 
{
  NSEnumerator *enumerator;
  NSArray      *result;
  id           obj, *objs;
  unsigned     cnt;
  
  objs = calloc([_flags count] + 2, sizeof(id));
  
  enumerator = [_flags objectEnumerator];
  cnt = 0;
  while ((obj = [enumerator nextObject])) {
    if ([obj length] == 0)
      continue;
    
    if (![[obj substringToIndex:1] isEqualToString:@"\\"])
      continue;

    objs[cnt] = [obj substringFromIndex:1];
    cnt++;
  }
  result = [NSArray arrayWithObjects:objs count:cnt];
  if (objs) free(objs);
  return result;
}

@end /* NGImap4ResponseNormalizer */
