/*
  Copyright (C) 2000-2004 SKYRIX Software AG

  This file is part of OpenGroupware.org.

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: WOPopUpButton.m,v 1.3 2004/04/30 12:12:24 helge Exp $

#include "WOPopUpButton.h"
#include "common.h"

#ifdef DEBUG
static int profElements  = -1;
static Class NSDateClass = Nil;

@interface WOContext(ComponentStackCount)
- (unsigned)componentStackCount;
@end

#endif

@implementation WOPopUpButton

- (id)initWithName:(NSString *)_name
  associations:(NSDictionary *)_config
  template:(WOElement *)_t
{
#if DEBUG
  if (profElements == -1) {
    profElements = [[[NSUserDefaults standardUserDefaults]
                                     objectForKey:@"WOProfileElements"]
                                     boolValue] ? 1 : 0;
  }
  if (NSDateClass == Nil)
    NSDateClass = [NSDate class];
#endif
  
  if ((self = [super initWithName:_name associations:_config template:_t])) {
    self->list              = OWGetProperty(_config, @"list");
    self->item              = OWGetProperty(_config, @"item");
    self->selection         = OWGetProperty(_config, @"selection");
    self->string            = OWGetProperty(_config, @"string");
    self->singleSelection   = OWGetProperty(_config, @"singleSelection");
    self->noSelectionString = OWGetProperty(_config, @"noSelectionString");

    // compatiblity
    if (self->noSelectionString == nil)
      self->noSelectionString = OWGetProperty(_config, @"nilString");
  }
  return self;
}

- (void)dealloc {
  RELEASE(self->noSelectionString);
  RELEASE(self->singleSelection);
  RELEASE(self->list);
  RELEASE(self->item);
  RELEASE(self->selection);
  RELEASE(self->string);
  [super dealloc];
}

/* handling the request */

- (void)takeValuesFromRequest:(WORequest *)_request
  inContext:(WOContext *)_ctx
{
  WOComponent *sComponent = [_ctx component];
    
  if (![self->disabled boolValueInComponent:[_ctx component]]) {
    NSString *formValue;
    
    formValue = [_request formValueForKey:OWFormElementName(self, _ctx)];
    //NSLog(@"%@: value=%@ ..", [self elementID], formValue);
    
    if ([self->value isValueSettable])
      [self->value setValue:formValue inComponent:sComponent];
    
    if (formValue) {
      NSArray *objects;
      id      object;
      
      objects = [self->list valueInComponent:sComponent];
      
      object = nil;
      if (self->value) {
        /* has a value binding, walk list to find object */
        unsigned i, toGo;

        for (i = 0, toGo = [objects count]; i < toGo; i++) {
          NSString *cv;
          
          object = [objects objectAtIndex:i];
          
          if ([self->item isValueSettable])
            [self->item setValue:object inComponent:sComponent];

          cv = [self->value stringValueInComponent:sComponent];
          
          if ([cv isEqualToString:formValue])
            break;
        }
      }
      else if (![formValue isEqualToString:@"$"]) {
        /* an index binding */
        int idx;
        
        idx = [formValue intValue];
        if (idx >= (int)[objects count]) {
          [[_ctx page] logWithFormat:@"popup-index %i out of range 0-%i",
                         idx, [objects count] - 1];
          object = nil;
        }
        else 
          object = [objects objectAtIndex:idx];
      }
      
      if ([self->selection isValueSettable]) {
        NSArray *sel;
        
        if ([self->item isValueSettable])
          [self->item setValue:object inComponent:sComponent];

        if (object) {
          if ([self->singleSelection boolValueInComponent:sComponent])
            sel = RETAIN(object);
          else
            sel = [[NSArray alloc] initWithObjects:object,nil];
        }
        else // nil item selected
          sel = nil;
          
        [self->selection setValue:sel inComponent:sComponent];
        RELEASE(sel); sel = nil;
      }
    }
    else {
      // nothing selected
      if ([self->item isValueSettable])
        [self->item setValue:nil inComponent:sComponent];
      if ([self->selection isValueSettable])
        [self->selection setValue:nil inComponent:sComponent];
    }
  }
}

- (void)appendOptionsToResponse:(WOResponse *)_response
  inContext:(WOContext *)_ctx
{
  WOComponent *sComponent = [_ctx component];
  BOOL     isSingle = NO;
  NSString *nilStr  = nil;
  NSArray  *array   = nil;
  id       selArray = nil;
  int      i, toGo;
#if DEBUG
  NSTimeInterval st = 0.0;
    
  if (profElements)
    st = [[NSDateClass date] timeIntervalSince1970];
#endif
    
  nilStr   = [self->noSelectionString stringValueInComponent:sComponent];
  isSingle = [self->singleSelection boolValueInComponent:sComponent];
  array    = [self->list            valueInComponent:sComponent];
  selArray = [self->selection       valueInComponent:sComponent];
  toGo     = [array count];
  
#if DEBUG
  if (profElements) {
    NSTimeInterval diff;
    int j;
    diff = [[NSDateClass date] timeIntervalSince1970] - st;
    if (diff > 0.001) {
      for (j = [_ctx componentStackCount]; j >= 0; j--)
        printf("  ");
      printf("PopUpOption[setup] %s: %0.3fs\n",
             [[_ctx elementID] cString], diff);
    }
  }
#endif
    
  if (nilStr) {
    WOResponse_AddCString(_response, "  <option value=\"$\">");
    WOResponse_AddHtmlString(_response, nilStr);
    WOResponse_AddCString(_response, "\n");
  }
  
  for (i = 0; i < toGo; i++) {
    NSString *v         = nil;
    NSString *displayV  = nil;
    id       object;
    BOOL     isSelected;
#if DEBUG
    NSTimeInterval st = 0.0;
    
    if (profElements)
      st = [[NSDateClass date] timeIntervalSince1970];
#endif
    
    object = [array objectAtIndex:i];
    
    if ([self->item isValueSettable])
      [self->item setValue:object inComponent:sComponent];

    if (isSingle)
      isSelected = selArray ? [selArray isEqual:object] : NO;
    else
      isSelected = selArray ? [selArray containsObject:object] : NO;

    v = self->value
      ? [self->value stringValueInComponent:sComponent]
      : [NSString stringWithFormat:@"%i", i];

    displayV = self->string
      ? [self->string stringValueInComponent:sComponent]
      : [object stringValue];

    if (displayV == nil) displayV = @"<nil>";
    
    WOResponse_AddCString(_response, "  <option value=\"");
    WOResponse_AddString(_response, v);
    WOResponse_AddCString(_response, 
			  isSelected ? 
			  "\" selected=\"selected\">" : "\">");
    WOResponse_AddHtmlString(_response, displayV);
    WOResponse_AddCString(_response, "</option>\n");
    
#if DEBUG
    if (profElements) {
      NSTimeInterval diff;
      int j;
      diff = [[NSDateClass date] timeIntervalSince1970] - st;
      if (diff > 0.001) {
#if 1
        for (j = [_ctx componentStackCount]; j >= 0; j--)
          printf("  ");
#endif
        printf("PopUpOption[%i] %s: %0.3fs\n", i,
               [[_ctx elementID] cString], diff);
      }
    }
#endif
  }
}

- (void)appendToResponse:(WOResponse *)_response inContext:(WOContext *)_ctx {
#if DEBUG
  NSTimeInterval st = 0.0;

  if (profElements)
    st = [[NSDateClass date] timeIntervalSince1970];
#endif
  
  WOResponse_AddCString(_response, "<select name=\"");
  [_response appendContentHTMLAttributeValue:OWFormElementName(self, _ctx)];
  WOResponse_AddChar(_response, '"');
  [self appendExtraAttributesToResponse:_response inContext:_ctx];
  if (self->otherTagString) {
    WOResponse_AddString(_response,
                         [self->otherTagString stringValueInComponent:
                              [_ctx component]]);
  }
  [_response appendContentString:@">\n"];

  [self appendOptionsToResponse:_response inContext:_ctx];

  WOResponse_AddCString(_response, "</select>");
  
#if DEBUG
  if (profElements) {
    NSTimeInterval diff;
    int i;
    diff = [[NSDateClass date] timeIntervalSince1970] - st;
    if (diff > 0.001) {
#if 1
      for (i = [_ctx componentStackCount]; i >= 0; i--)
        printf("  ");
#endif
      printf("PopUpButton %s: %0.3fs\n",
             [[_ctx elementID] cString], diff);
    }
  }
#endif
}

/* description */

- (NSString *)associationDescription {
  NSMutableString *str;
  
  str = [NSMutableString stringWithCapacity:256];
  [str appendString:[super associationDescription]];

  if (self->list)      [str appendFormat:@" list=%@",      self->list];
  if (self->item)      [str appendFormat:@" item=%@",      self->item];
  if (self->selection) [str appendFormat:@" selection=%@", self->selection];
  if (self->string)    [str appendFormat:@" string=%@",    self->string];
  if (self->noSelectionString)
    [str appendFormat:@" noselection=%@", self->noSelectionString];
  if (self->singleSelection)
    [str appendFormat:@" singleSelection=%@", self->singleSelection];

  return str;
}

@end /* WOPopUpButton */
