/*
  Copyright (C) 2000-2003 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: SoPageInvocation.m,v 1.3 2003/09/06 21:50:28 znek Exp $

#include "SoPageInvocation.h"
#include "SoClassSecurityInfo.h"
#include "SoProduct.h"
#include "WOContext+SoObjects.h"
#include <NGObjWeb/WOComponent.h>
#include <NGObjWeb/WOContext.h>
#include <NGObjWeb/WOSession.h>
#include <NGObjWeb/WOResponse.h>
#include <NGObjWeb/WORequest.h>
#include <NGObjWeb/WOResourceManager.h>
#include <NGObjWeb/WOApplication.h>
#include "common.h"

@interface WOComponent(DirectActionExtensions)
/* this is defined in WOPageRequestHandler */
- (id<WOActionResults>)performActionNamed:(NSString *)_actionName;
- (id<WOActionResults>)defaultAction;
- (void)setResourceManager:(WOResourceManager *)_rm;
@end

@interface WOContext(UsedPrivates)
- (void)setPage:(WOComponent *)_page;
@end

@implementation SoPageInvocation

static int debugOn = 0;

+ (void)initialize {
  static BOOL didInit = NO;
  if (!didInit) {
    NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
    didInit = YES;
    
    debugOn = [ud boolForKey:@"SoPageInvocationDebugEnabled"] ? 1 : 0;
  }
}

- (id)initWithPageName:(NSString *)_pageName actionName:(NSString *)_action {
  if ((self = [super init])) {
    self->pageName   = [_pageName copy];
    self->actionName = [_action   copy];
  }
  return self;
}
- (id)initWithPageName:(NSString *)_pageName {
  return [self initWithPageName:_pageName actionName:nil];
}
- (id)init {
  return [self initWithPageName:nil];
}
- (id)initWithPageName:(NSString *)_pageName actionName:(NSString *)_action
  product:(SoProduct *)_product
{
  if ((self = [self initWithPageName:_pageName actionName:_action])) {
    self->product = _product;
  }
  return self;
}

- (void)dealloc {
  [self->page       release];
  [self->object     release];
  [self->pageName   release];
  [self->actionName release];
  [super dealloc];
}

/* containment */

- (void)detachFromContainer {
  self->product = nil;
}
- (id)container {
  return self->product;
}
- (NSString *)nameInContainer {
  /* could ask product */
  return nil;
}

/* accessors */

- (NSString *)pageName {
  return self->pageName;
}
- (NSString *)actionName {
  return self->actionName;
}

/* page construction */

- (WOComponent *)instantiatePageInContext:(id)_ctx {
  WOResourceManager *rm;
  WOComponent *lPage;
  NSArray     *languages;
  
  if (debugOn) {
    [self debugWithFormat:@"instantiate page: %@", self->pageName];
    if (self->product == nil)
      [self debugWithFormat:@"  no product is set."];
  }
  
  /* lookup available resource manager (product,component,app) */
  
  if ((rm = [self->product resourceManager]) == nil) {
    if ((rm = [[_ctx component] resourceManager]) == nil) {
      rm = [[_ctx application] resourceManager];
      if (debugOn) [self debugWithFormat:@"   app-rm: %@", rm];
    }
    else
      if (debugOn) [self debugWithFormat:@"   component-rm: %@", rm];
  }
  else
    if (debugOn) [self debugWithFormat:@"   product-rm: %@", rm];
  
  /* determine language */
  
  languages = [_ctx hasSession]
    ? [(WOSession *)[_ctx session] languages]
    : [[(id <WOPageGenerationContext>)_ctx request] browserLanguages];

  /* instantiate */
  
  lPage = [rm pageWithName:self->pageName languages:languages];
  [lPage ensureAwakeInContext:_ctx];
  [lPage setResourceManager:rm];
  
  if (debugOn) [self debugWithFormat:@"   page: %@", lPage];
  
  return lPage;
}

/* invocation */

- (BOOL)isCallable {
  return YES;
}
- (id)clientObject {
  return self->object;
}

- (id)callOnObject:(id)_client inContext:(id)_ctx {
  WOComponent *lPage;
  
  if (self->object != _client) {
    /* rebind */
    return [[self bindToObject:_client inContext:_ctx]
	          callOnObject:_client inContext:_ctx];
  }
  
  if ((lPage = self->page) == nil)
    lPage = [self instantiatePageInContext:_ctx];
  
  if (lPage == nil) {
    [self logWithFormat:@"found no page named '%@' for call !", 
	  self->pageName];
    return nil;
  }
  
  /* make page the "request" page */
  [_ctx setPage:lPage];
  
  [lPage setClientObject:_client];
  
  if ([lPage shouldTakeValuesFromRequest:[(id <WOPageGenerationContext>)_ctx request] inContext:_ctx]) {
    [[_ctx application] takeValuesFromRequest:[(id <WOPageGenerationContext>)_ctx request] inContext:_ctx];
  }
  
  if (self->actionName) {
    if (debugOn) {
      [self debugWithFormat:@"  performing action %@ on page: %@", 
	      self->actionName, lPage];
    }
    return [lPage performActionNamed:self->actionName];
  }
  else {
    if (debugOn)
      [self debugWithFormat:@"  performing default action on page: %@", lPage];
    return [lPage defaultAction];
  }
}

/* bindings */

- (BOOL)isBound {
  return self->object != nil ? YES : NO;
}

- (id)bindToObject:(id)_object inContext:(id)_ctx {
  SoPageInvocation *inv;
  
  if (_object == nil) return nil;
  
  // TODO: clean up this section, a bit hackish
  inv = [[SoPageInvocation alloc] initWithPageName:self->pageName];
  inv = [inv autorelease];
  
  inv->product    = self->product; // non-owned (cannot be detached !!!)
  inv->object     = [_object retain];
  inv->actionName = [self->actionName copy];
  inv->page = [[inv instantiatePageInContext:_ctx] retain];
  
  if (inv->page == nil) {
    [self logWithFormat:@"ERROR: did not find page method '%@'", 
	    self->pageName];
    return nil;
  }
  
  return inv;
}

/* delivering as content (can happen in DAV !) */

- (void)appendToResponse:(WOResponse *)_r inContext:(WOContext *)_ctx {
  [_r appendContentString:@"native component method: "];
  [_r appendContentHTMLString:[self description]];
}

/* description */

- (NSString *)description {
  NSMutableString *ms;

  ms = [NSMutableString stringWithCapacity:64];
  [ms appendFormat:@"<0x%08X[%@]:", self, NSStringFromClass([self class])];
  
  if (self->pageName)   [ms appendFormat:@" page=%@",   self->pageName];
  if (self->actionName) [ms appendFormat:@" action=%@", self->actionName];
  
  if (self->object) [ms appendString:@" bound"];
  if (self->page)   [ms appendString:@" instantiated"];
  
  if (self->product) [ms appendFormat:@" product=%@", self->product];
  
  [ms appendString:@">"];
  return ms;
}

@end /* SoPageInvocation */

@implementation SoPageInvocation(Logging)

- (NSString *)loggingPrefix {
  return [NSString stringWithFormat:@"[so-page 0x%08X %@]", 
		     self, self->pageName];
}
- (BOOL)isDebuggingEnabled {
  return debugOn ? YES : NO;
}

@end /* SoPageInvocation(Logging) */
