/*
  Copyright (C) 2000-2004 SKYRIX Software AG

  This file is part of OpenGroupware.org.

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: SoProductRegistry.m,v 1.6 2004/03/08 00:41:57 helge Exp $

#include "SoProductRegistry.h"
#include "SoProduct.h"
#include "SoObject.h"
#include "SoClassSecurityInfo.h"
#include "common.h"
#include <NGObjWeb/WOResponse.h>
#include <NGObjWeb/WOContext.h>

@implementation SoProductRegistry

static int debugOn = 0;

+ (void)initialize {
  static BOOL didInit = NO;
  if (!didInit) {
    NSUserDefaults *ud = [NSUserDefaults standardUserDefaults];
    didInit = YES;
    debugOn = [ud boolForKey:@"SoProductRegistryDebugEnabled"] ? 1 : 0;
  }
}

+ (id)sharedProductRegistry {
  static SoProductRegistry *reg = nil; // THREAD
  if (reg == nil)
    reg = [[SoProductRegistry alloc] init];
  return reg;
}

- (id)init {
  if ((self = [super init])) {
    [self scanForAvailableProducts];
  }
  return self;
}

- (void)dealloc {
  [self->products release];
  [super dealloc];
}

/* operations */

- (NSFileManager *)fileManager {
  return [NSFileManager defaultManager];
}

- (void)registerProductBundle:(NSBundle *)_bundle {
  NSString  *productName;
  SoProduct *product;
  NSString  *manifest;
  
  if (_bundle != [NSBundle mainBundle]) {
    productName =
      [[[_bundle bundlePath] lastPathComponent] stringByDeletingPathExtension];
  }
  else
    productName = @"MAIN";
  
  if ((product = [self->products objectForKey:productName])) {
    [self debugWithFormat:@"  product '%@' already registered.", productName];
    [product reloadIfPossible];
  }
  
  manifest = [_bundle pathForResource:@"product" ofType:@"plist"];
  if ([manifest length] == 0) {
    if ([productName isEqualToString:@"MAIN"])
      [self debugWithFormat:@"  main bundle has no manifest."];
    return;
  }
  
  [self debugWithFormat:@"register product bundle: '%@'", 
	  [_bundle bundlePath]];
  
  if (self->products == nil)
    self->products = [[NSMutableDictionary alloc] initWithCapacity:32];
  
  [self debugWithFormat:@"  register as product: %@", productName];
  
  if ((product = [[SoProduct alloc] initWithBundle:_bundle]) == nil) {
    [self debugWithFormat:@"  could init product from bundle: %@", _bundle];
    return;
  }
  
  [self->products setObject:product forKey:productName];
  [product release];
}

- (void)registerProductAtPath:(NSString *)_path {
  NSBundle *bundle;
  
  if ((bundle = [NGBundle bundleWithPath:_path]) == nil) {
    [self logWithFormat:@"couldn't init bundle object for path: %@", _path];
    return;
  }
  [self registerProductBundle:bundle];
}

- (void)scanForProductsInDirectory:(NSString *)_path {
  NSFileManager *fm;
  NSEnumerator  *pathes;
  NSString      *lPath;
  
  fm = [self fileManager];
  pathes = [[fm directoryContentsAtPath:_path] objectEnumerator];
  while ((lPath = [pathes nextObject])) {
    BOOL isDir;
    
    lPath = [_path stringByAppendingPathComponent:lPath];
    
    if (![fm fileExistsAtPath:lPath isDirectory:&isDir])
      continue;
    if (!isDir)
      continue;
    
    [self registerProductAtPath:lPath];
  }
}

- (void)scanForAvailableProducts {
  NSFileManager *fm;
  NSProcessInfo *pi;
  NSArray  *pathes;
  NSBundle *bundle;
  unsigned i;

  /* scan mail bundle & frameworks */
  
  if ((bundle = [NSBundle mainBundle]))
    [self registerProductBundle:bundle];
  else
    NSLog(@"%s: missing main bundle ...", __PRETTY_FUNCTION__);
  
  pathes = [NSBundle allFrameworks];
  for (i = 0; i < [pathes count]; i++)
    [self registerProductBundle:[pathes objectAtIndex:i]];

  pathes = [NSBundle allBundles];
  for (i = 0; i < [pathes count]; i++)
    [self registerProductBundle:[pathes objectAtIndex:i]];
  
  /* scan library pathes */
  
  fm = [NSFileManager defaultManager];
  pi = [NSProcessInfo processInfo];
  
#if COCOA_Foundation_LIBRARY
  /* 
     TODO: (like COMPILE_FOR_GNUSTEP)
     This should actually check whether we are compiling in the
     GNUstep environment since this modifies the location of bundles.
  */
  pathes = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,
					       NSAllDomainsMask,
					       YES);
#else
  pathes = [[pi environment] objectForKey:@"GNUSTEP_PATHPREFIX_LIST"];
  if (pathes == nil)
    pathes = [[pi environment] objectForKey:@"GNUSTEP_PATHLIST"];
  
  pathes = [[pathes stringValue] componentsSeparatedByString:@":"];
#endif

  if ([pathes count] == 0) {
    [self debugWithFormat:@"found no product pathes."];
    return;
  }
  
  [self debugWithFormat:@"scanning for products ..."];
  
  for (i = 0; i < [pathes count]; i++) {
    NSString *lPath;
    BOOL     isDir;
    
    lPath = [pathes objectAtIndex:i];
#if !COCOA_Foundation_LIBRARY
    lPath = [lPath stringByAppendingPathComponent:@"Library"];
#endif
    lPath = [lPath stringByAppendingPathComponent:@"SoProducts"];
    
    if (![fm fileExistsAtPath:lPath isDirectory:&isDir])
      continue;
    if (!isDir)
      continue;
    
    [self debugWithFormat:@"  directory %@", lPath];
    [self scanForProductsInDirectory:lPath];
  }
  
  [self debugWithFormat:
	  @"finished scan for products (%i products registered).",
	  [self->products count]];
}

/* registering products */

- (BOOL)loadProductNamed:(NSString *)_name {
  SoProduct    *product;
  NSEnumerator *requiredProducts;
  NSString     *rqname;
  
  if ((product = [self->products objectForKey:_name]) == nil) {
    [self debugWithFormat:@"did not find product: %@", _name];
    return NO;
  }
  
  /* load dependencies (TODO: should detect cycles) */
  requiredProducts = [[product requiredProducts] objectEnumerator];
  while ((rqname = [requiredProducts nextObject])) {
    if (![self loadProductNamed:rqname]) {
      if ([rqname isEqualToString:@"MAIN"]) continue;
      [self logWithFormat:@"ERROR: failed to load product %@ required by %@.",
	      rqname, _name];
      return NO;
    }
  }
  
  return [product load];
}

- (BOOL)loadAllProducts {
  NSEnumerator *e;
  NSString *p;
  
  e = [self->products keyEnumerator];
  while ((p = [e nextObject])) {
    if (![self loadProductNamed:p])
      [self logWithFormat:@"could not load product: %@", p];
  }
  return YES;
}

/* lookup products */

- (NSArray *)registeredProductNames {
  return [self->products allKeys];
}
- (SoProduct *)productWithName:(NSString *)_name {
  return [self->products objectForKey:_name];
}

/* bundle */

- (SoProduct *)productForBundle:(NSBundle *)_bundle {
  /* TODO: add a registry based on path ... */
  NSString  *pname;
  SoProduct *product;
  
  pname =
    [[[_bundle bundlePath] lastPathComponent] stringByDeletingPathExtension];
  
  if ((product = [self productWithName:pname]))
    return product;
  
  [self logWithFormat:
          @"product '%@' not yet registered, attempting to load ...", pname];
  if (![self loadProductNamed:pname])
    return nil;
  return [self productWithName:pname];
}

/* product registry as a SoObject */

- (NSArray *)allKeys {
  return [self registeredProductNames];
}

- (BOOL)hasName:(NSString *)_key inContext:(id)_ctx {
  if ([self->products objectForKey:_key])
    return YES;
  return [super hasName:_key inContext:_ctx];
}

- (id)lookupName:(NSString *)_key inContext:(id)_ctx acquire:(BOOL)_flag {
  SoProduct *product;
  
  if ((product = [self productWithName:_key]))
    return product;
  
  return [super lookupName:_key inContext:_ctx acquire:_flag];
}

- (NSArray *)toOneRelationshipKeys {
  NSMutableSet *ma;
  id root;

  if ((root = [super toOneRelationshipKeys]) == nil)
    return [self->products allKeys];
  
  ma = [[NSMutableSet alloc] initWithArray:root];
  [ma addObjectsFromArray:[self->products allKeys]];
  root = [ma allObjects];
  [ma release];
  return root;
}

/* debugging */

- (NSString *)loggingPrefix {
  return @"[so-product-registry]";
}
- (BOOL)isDebuggingEnabled {
  return debugOn ? YES : NO;
}

/* web representation */

- (void)appendToResponse:(WOResponse *)_r inContext:(WOContext *)_ctx {
  NSEnumerator *e;
  NSString *name;
  
  [_r appendContentString:@"<h3>SOPE Product Registry</h3>"];
  
  e = [[self toOneRelationshipKeys] objectEnumerator];
  while ((name = [e nextObject])) {
    [_r appendContentString:@"<li><a href=\""];
    [_r appendContentHTMLAttributeValue:name];
    [_r appendContentString:@"\">"];
    [_r appendContentHTMLString:name];
    [_r appendContentString:@"</a></li>"];
  }    
}

@end /* SoProductRegistry */
