/* 
   EOAdaptorChannel.m

   Copyright (C) 1996 Free Software Foundation, Inc.

   Author: Ovidiu Predescu <ovidiu@bx.logicnet.ro>
   Date: October 1996

   This file is part of the GNUstep Database Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#import "common.h"
#import "EOAdaptorChannel.h"
#import "EOAttribute.h"
#import "EOAdaptor.h"
#import "EOAdaptorContext.h"
#import "EOSQLExpression.h"
#import "EOFExceptions.h"

@interface EOAdaptorChannel(Internals)
- (NSArray *)_sortAttributesForSelectExpression:(NSArray *)_attrs;
@end /* EOAdaptorChannel(Internals) */

@implementation EOAdaptorChannel

+ (NSCalendarDate*)dateForAttribute:(EOAttribute*)attr 
  year:(int)year month:(unsigned)month day:(unsigned)day 
  hour:(unsigned)hour minute:(unsigned)minute second:(unsigned)second 
  zone:(NSZone*)zone
{
    NSTimeZone     *serverTimeZone = [attr serverTimeZone];
    NSTimeZone     *clientTimeZone = [attr clientTimeZone];
    NSCalendarDate *date;
    NSString       *fmt;

    if (serverTimeZone == nil) serverTimeZone = [NSTimeZone localTimeZone];
    if (clientTimeZone == nil) clientTimeZone = [NSTimeZone localTimeZone];
    
    date = AUTORELEASE([[NSCalendarDate allocWithZone:zone]
                           initWithYear:year month:month day:day hour:hour
                           minute:minute second:second timeZone:serverTimeZone]);
    [date setTimeZone:clientTimeZone];
    fmt = [attr calendarFormat];
    [date setCalendarFormat:fmt ? fmt : [EOAttribute defaultCalendarFormat]];
    return date;
}

- (id)initWithAdaptorContext:(EOAdaptorContext*)_adaptorContext {
    ASSIGN(self->adaptorContext, _adaptorContext);
    
    NS_DURING
        [self->adaptorContext channelDidInit:self];
    NS_HANDLER {
      AUTORELEASE(self);
      [localException raise];
    }
    NS_ENDHANDLER;
    
    return self;
}

- (void)dealloc {
    [self->adaptorContext channelWillDealloc:self];
    RELEASE(self->adaptorContext);
    [super dealloc];
}

- (BOOL)openChannel {
    if(self->isOpen)
        return NO;
    self->isOpen = YES;

    return YES;
}

- (void)closeChannel {
    if([self isFetchInProgress])
        [self cancelFetch];
    self->isOpen = NO;
}

- (BOOL)insertRow:(NSDictionary *)row forEntity:(EOEntity *)entity {
    EOSQLExpression     *sqlexpr;
    NSMutableDictionary *mrow = (id)row;

    if (!isOpen)
        [[ChannelIsNotOpenedException new] raise];

    if((row == nil) || (entity == nil)) {
      [NSException raise:NSInvalidArgumentException
		   format:@"row and entity arguments for insertRow:forEntity:"
                           @"must not be the nil object"];
    }

    if([self isFetchInProgress])
        [[AdaptorIsFetchingException exceptionWithAdaptor:self] raise];

    if(![adaptorContext transactionNestingLevel])
        [[NoTransactionInProgressException exceptionWithAdaptor:self] raise];

    if(delegateRespondsTo.willInsertRow) {
        EODelegateResponse response;

        mrow = AUTORELEASE([row mutableCopyWithZone:[row zone]]);
        response = [delegate adaptorChannel:self
                             willInsertRow:mrow
                             forEntity:entity];
        if(response == EODelegateRejects)
            return NO;
        else if(response == EODelegateOverrides)
            return YES;
    }

    sqlexpr = [[[adaptorContext adaptor]
                                expressionClass]
                                insertExpressionForRow:mrow
                                entity:entity
                                channel:self];
    
    if(![self evaluateExpression:[sqlexpr expressionValueForContext:nil]])
        return NO;

    if(delegateRespondsTo.didInsertRow)
        [delegate adaptorChannel:self didInsertRow:mrow forEntity:entity];

    return YES;
}

- (BOOL)updateRow:(NSDictionary*)row
  describedByQualifier:(EOSQLQualifier*)qualifier
{
    EOSQLExpression     *sqlexpr = nil;
    NSMutableDictionary *mrow    = (id)row;

    if (!isOpen)
        [[ChannelIsNotOpenedException new] raise];

    if (row == nil) {
      [NSException raise:NSInvalidArgumentException
		   format:@"row argument for updateRow:describedByQualifier: "
                           @"must not be the nil object"];
    }

    if([self isFetchInProgress])
        [[AdaptorIsFetchingException exceptionWithAdaptor:self] raise];

    if(![adaptorContext transactionNestingLevel])
        [[NoTransactionInProgressException exceptionWithAdaptor:self] raise];

    if(delegateRespondsTo.willUpdateRow) {
        EODelegateResponse response;

        mrow = AUTORELEASE([row mutableCopyWithZone:[row zone]]);
        response = [delegate adaptorChannel:self
                             willUpdateRow:mrow
                             describedByQualifier:qualifier];
        if(response == EODelegateRejects)
            return NO;
        else if(response == EODelegateOverrides)
            return YES;
    }

    sqlexpr = [[[adaptorContext adaptor]
                                expressionClass]
                                updateExpressionForRow:mrow
                                qualifier:qualifier
                                channel:self];
    
    if(![self evaluateExpression:[sqlexpr expressionValueForContext:nil]])
        return NO;

    if(delegateRespondsTo.didUpdateRow) {
        [delegate adaptorChannel:self
                  didUpdateRow:mrow
                  describedByQualifier:qualifier];
    }
    return YES;
}

- (BOOL)deleteRowsDescribedByQualifier:(EOSQLQualifier*)qualifier {
    EOSQLExpression *sqlexpr = nil;

    if (!isOpen)
      [[ChannelIsNotOpenedException new] raise];

    if([self isFetchInProgress])
      [[AdaptorIsFetchingException exceptionWithAdaptor:self] raise];

    if(![adaptorContext transactionNestingLevel])
      [[NoTransactionInProgressException exceptionWithAdaptor:self] raise];
    
    if(delegateRespondsTo.willDeleteRows) {
        EODelegateResponse response
                = [delegate adaptorChannel:self
                            willDeleteRowsDescribedByQualifier:qualifier];
        if(response == EODelegateRejects)
            return NO;
        else if(response == EODelegateOverrides)
            return YES;
    }

    sqlexpr = [[[adaptorContext adaptor]
                                expressionClass]
                                deleteExpressionWithQualifier:qualifier
                                channel:self];
    
    if(![self evaluateExpression:[sqlexpr expressionValueForContext:nil]])
        return NO;

    if(delegateRespondsTo.didDeleteRows)
        [delegate adaptorChannel:self
                  didDeleteRowsDescribedByQualifier:qualifier];

    return YES;
}

- (BOOL)selectAttributes:(NSArray *)attributes
  describedByQualifier:(EOSQLQualifier *)qualifier
  fetchOrder:(NSArray *)fetchOrder
  lock:(BOOL)lockFlag
{
    EOSQLExpression *sqlexpr = nil;
    NSMutableArray  *mattrs  = (NSMutableArray *)attributes;
    NSMutableArray  *mfetch  = (NSMutableArray *)fetchOrder;

    if (!isOpen)
      [[ChannelIsNotOpenedException new] raise];

    if (attributes == nil) {
      [NSException raise:NSInvalidArgumentException
		   format:@"attributes argument for selectAttributes:"
		   @"describedByQualifier:fetchOrder:lock: "
		   @"must not be the nil object"];
    }

    if([self isFetchInProgress])
      [[AdaptorIsFetchingException exceptionWithAdaptor:self] raise];

    if(![adaptorContext transactionNestingLevel])
      [[NoTransactionInProgressException exceptionWithAdaptor:self] raise];

    if(delegateRespondsTo.willSelectAttributes) {
        EODelegateResponse response;
        
        mattrs = [[attributes mutableCopy] autorelease];
        mfetch = [[fetchOrder mutableCopy] autorelease];

        response = [delegate adaptorChannel:self
                             willSelectAttributes:mattrs
                             describedByQualifier:qualifier
                             fetchOrder:mfetch
                             lock:lockFlag];
        if (response == EODelegateRejects)
            return NO;
	if (response == EODelegateOverrides)
            return YES;
    }

#if 0
#warning DEBUG LOG, REMOVE!
    [self logWithFormat:@"fetch qualifier: %@", qualifier];
#endif

    sqlexpr = [[[adaptorContext adaptor]
                                expressionClass]
                                selectExpressionForAttributes:attributes
                                lock:lockFlag
                                qualifier:qualifier
                                fetchOrder:fetchOrder
                                channel:self];
    
    if (![self evaluateExpression:[sqlexpr expressionValueForContext:nil]])
        return NO;

    if (delegateRespondsTo.didSelectAttributes) {
        [delegate adaptorChannel:self
                  didSelectAttributes:mattrs
                  describedByQualifier:qualifier
                  fetchOrder:mfetch
                  lock:lockFlag];
    }

    return YES;
}

- (NSArray*)describeResults {
  [self subclassResponsibility:_cmd];
  return nil;
}

- (NSMutableDictionary *)fetchAttributes:(NSArray *)attributes
  withZone:(NSZone *)_zone
{
    NSMutableDictionary *row = nil;

    if (!self->isOpen)
      [[ChannelIsNotOpenedException new] raise];

    if (_zone == NULL)
        _zone = NSDefaultMallocZone();

    if(![self isFetchInProgress])
      [[AdaptorIsNotFetchingException exceptionWithAdaptor:self] raise];

    if(delegateRespondsTo.willFetchAttributes) {
        row = [delegate adaptorChannel:self
                        willFetchAttributes:attributes
                        withZone:_zone];
    }

    /* NOTE: primaryFetchAttributes:withZone: have to set the isFetchInProgress
       status */
    if(row == nil)
        row = [self primaryFetchAttributes:attributes withZone:_zone];

    if(row) {
        if(delegateRespondsTo.didFetchAttributes)
            row = [delegate adaptorChannel:self
                            didFetchAttributes:row
                            withZone:_zone];
        if(delegateRespondsTo.didChangeResultSet)
            [delegate adaptorChannelDidChangeResultSet:self];
    }
    else {
        /* Do not set here the isFetchInProgress status since only subclasses
           can know whether there are more SELECT commands to be executed.
           Setting the status here to NO will overwrite the adaptor subclass
           intention. */
        if(delegateRespondsTo.didFinishFetching)
            [delegate adaptorChannelDidFinishFetching:self];
    }
    return row;
}

- (BOOL)isFetchInProgress {
    return self->isFetchInProgress;
}

- (void)cancelFetch {
    if (!isOpen)
      [[ChannelIsNotOpenedException new] raise];

    self->isFetchInProgress = NO;
}

- (NSMutableDictionary *)dictionaryWithObjects:(id *)objects 
  forAttributes:(NSArray *)attributes zone:(NSZone *)zone
{
    [self notImplemented:_cmd];
    return nil;
}

- (NSMutableDictionary*)primaryFetchAttributes:(NSArray*)attributes 
  withZone:(NSZone*)zone
{
    [self subclassResponsibility:_cmd];
    return nil;
}

- (BOOL)evaluateExpression:(NSString *)anExpression {
    [self subclassResponsibility:_cmd];
    return NO;
}

- (EOAdaptorContext*)adaptorContext {
    return self->adaptorContext;
}

- (EOModel *)describeModelWithTableNames:(NSArray*)tableNames {
    [self subclassResponsibility:_cmd];
    return nil;
}

- (NSArray *)describeTableNames {
    [self subclassResponsibility:_cmd];
    return nil;
}

- (BOOL)readTypesForEntity:(EOEntity*)anEntity {
    [self subclassResponsibility:_cmd];
    return NO;
}

- (BOOL)readTypeForAttribute:(EOAttribute*)anAttribute {
    [self subclassResponsibility:_cmd];
    return NO;
}

// delegate

- (id)delegate {
    return self->delegate;
}

- (void)setDelegate:(id)_delegate {
    self->delegate = _delegate;

    delegateRespondsTo.willInsertRow = 
        [delegate respondsToSelector:
                @selector(adaptorChannel:willInsertRow:forEntity:)];
    delegateRespondsTo.didInsertRow = 
        [delegate respondsToSelector:
                @selector(adaptorChannel:didInsertRow:forEntity:)];
    delegateRespondsTo.willUpdateRow =
        [delegate respondsToSelector:
                @selector(adaptorChannel:willUpdateRow:describedByQualifier:)];
    delegateRespondsTo.didUpdateRow =
        [delegate respondsToSelector:
                @selector(adaptorChannel:didUpdateRow:describedByQualifier:)];
    delegateRespondsTo.willDeleteRows =
        [delegate respondsToSelector:
                @selector(adaptorChannel:willDeleteRowsDescribedByQualifier:)];
    delegateRespondsTo.didDeleteRows =
        [delegate respondsToSelector:
                @selector(adaptorChannel:didDeleteRowsDescribedByQualifier:)];
    delegateRespondsTo.willSelectAttributes =
        [delegate respondsToSelector:
                @selector(adaptorChannel:willSelectAttributes:
                          describedByQualifier:fetchOrder:lock:)];
    delegateRespondsTo.didSelectAttributes =
        [delegate respondsToSelector:
                @selector(adaptorChannel:didSelectAttributes:
                          describedByQualifier:fetchOrder:lock:)];
    delegateRespondsTo.willFetchAttributes =
        [delegate respondsToSelector:
                @selector(adaptorChannel:willFetchAttributes:withZone:)];
    delegateRespondsTo.didFetchAttributes =
        [delegate respondsToSelector:
                @selector(adaptorChannel:didFetchAttributes:withZone:)];
    delegateRespondsTo.didChangeResultSet =
        [delegate respondsToSelector:
                @selector(adaptorChannelDidChangeResultSet:)];
    delegateRespondsTo.didFinishFetching =
        [delegate respondsToSelector:
                @selector(adaptorChannelDidFinishFetching:)];
    delegateRespondsTo.willEvaluateExpression =
        [delegate respondsToSelector:
                @selector(adaptorChannel:willEvaluateExpression:)];
    delegateRespondsTo.didEvaluateExpression =
        [delegate respondsToSelector:
                @selector(adaptorChannel:didEvaluateExpression:)];
}

- (void)setDebugEnabled:(BOOL)flag {
    self->debugEnabled = flag;
}

- (BOOL)isDebugEnabled {
    return self->debugEnabled;
}
- (BOOL)isOpen {
    return self->isOpen;
}

@end /* EOAdaptorChannel */

#import <EOControl/EOFetchSpecification.h>

@implementation EOAdaptorChannel(PrimaryKeyGeneration) // new in EOF2

- (NSDictionary *)primaryKeyForNewRowWithEntity:(EOEntity *)_entity {
  return nil;
}

@end /* EOAdaptorChannel(PrimaryKeyGeneration) */

@implementation EOAdaptorChannel(EOF2Additions)

- (void)selectAttributes:(NSArray *)_attributes
  fetchSpecification:(EOFetchSpecification *)_fspec
  lock:(BOOL)_flag
  entity:(EOEntity *)_entity
{
  EOSQLQualifier *q;
  BOOL isOk;
  
  q = (EOSQLQualifier *)[_fspec qualifier];
  
  isOk = [self selectAttributes:_attributes
               describedByQualifier:q
               fetchOrder:[_fspec sortOrderings]
               lock:_flag];
  if (!isOk) {
    [NSException raise:@"Select failed"
                 format:@"could not select attributes for fetch spec"];
  }
}

- (void)setAttributesToFetch:(NSArray *)_attributes {
  [self notImplemented:_cmd];
}
- (NSArray *)attributesToFetch {
  NSLog(@"ERROR(%s): subclasses must override this method!", 
	__PRETTY_FUNCTION__);
  return nil;
}

- (NSMutableDictionary *)fetchRowWithZone:(NSZone *)_zone {
  return [self fetchAttributes:[self attributesToFetch] withZone:_zone];
}

@end /* EOAdaptorChannel(EOF2Additions) */

@implementation EOAdaptorChannel(Internals)

- (NSArray *)_sortAttributesForSelectExpression:(NSArray *)_attrs {
  return _attrs;
}

@end /* EOAdaptorChannel(Internals) */
