/*
  Copyright (C) 2000-2003 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: SkyDisplayPreferences.m,v 1.3 2003/09/03 14:23:39 martin Exp $

#include <OGoFoundation/LSWContentPage.h>
#include <OGoFoundation/LSWNotifications.h>
#include "common.h"

@interface SkyDisplayPreferences : LSWContentPage {
  id             account;
  id             item;
  NSUserDefaults *defaults;
  BOOL           isRoot;
  NSDictionary   *availableDockablePages;

  /* values */
  NSArray        *dockablePages;
  NSMutableArray *dockedPages;
  NSMutableArray *undockedPages;
  
  BOOL           defaultButtonsLeft;
  BOOL           hideInactiveButtons;
  BOOL           buttonTextMode;
  BOOL           dockTextMode;
  BOOL           dockShowBookmarks;
  BOOL           dockShowDesktop;
  int            maxClipCount;
  int            maxNavLabelLength;
  NSString       *language;
  NSString       *timezone;
  BOOL           showAMPMDates;

  /* access */
  BOOL           isDockablePagesEditable;
  BOOL           isDockShowBookmarksEditable;
  BOOL           isDockShowDesktopEditable;
  BOOL           isLanguageEditable;
  BOOL           isTimezoneEditable;
}

- (void)setItem:(id)_item;
- (id)item;

@end /* SkyDisplayPreferences */

@implementation SkyDisplayPreferences

- (id)init {
  if ((self = [super init])) {
    NGBundleManager *bm;
    WOSession       *sn;
    BOOL            isExtraAccount, isRootAccount;
    NSArray         *tmp;
    
    bm = [NGBundleManager defaultBundleManager];
    sn = [self session];

    /* get type of current account */

    isRootAccount  = [sn activeAccountIsRoot];
    isExtraAccount = [[[sn activeAccount]
                           valueForKey:@"isExtraAccount"] boolValue];

    /* query bundle manager for all available dockables */

    if ([(tmp = [bm providedResourcesOfType:@"DockablePages"]) count] > 0) {
      NSMutableDictionary *dict;
      NSMutableSet        *uniquer;
      int                 count, i;

      count   = [tmp count];
      dict    = [NSMutableDictionary dictionaryWithCapacity:count];
      uniquer = [NSMutableSet setWithCapacity:count];

      for (i = 0; i < [tmp count]; i++) {
        NSDictionary *cfgEntry;
        NSBundle     *bundle;
        NSString     *componentName, *label, *icon, *pageName;;

        cfgEntry = [tmp objectAtIndex:i];
        pageName = [cfgEntry objectForKey:@"name"];

        if (pageName == nil) {
          [self logWithFormat:@"missing pagename in bundle config %@",
                cfgEntry];
          continue;
        }

        if ([uniquer containsObject:pageName])
          /* already found a bundle providing that resource */
          continue;

        [uniquer addObject:pageName];

        bundle = [bm bundleProvidingResource:pageName ofType:@"DockablePages"];
        if (bundle == nil) {
          [self logWithFormat:@"did not find dockable page %@", pageName];
          continue;
        }

        /* refetch config entry, so that it is the right one .. */

        cfgEntry = [bundle configForResource:pageName ofType:@"DockablePages"];
        if (cfgEntry == nil) {
          [self logWithFormat:@"missing configuration of dockable page %@",
                  pageName];
          continue;
        }
        componentName = [cfgEntry objectForKey:@"component"];
        label         = [cfgEntry objectForKey:@"labelKey"];
        icon          = [cfgEntry objectForKey:@"listicon"];

        /* check whether page is visible for root only */

        if ([[cfgEntry objectForKey:@"onlyRoot"] boolValue]) {
          if (!isRootAccount)
            continue;
        }

        /* check whether the page is visible for extra accounts */
    
        if (isExtraAccount) {
          if (![[cfgEntry objectForKey:@"allowedForExtraAccounts"]
                          boolValue]) {
            NSLog(@"page %@ is not allowed for extra accounts !", pageName);
            continue;
          }
        }
        if (componentName == nil) {
          [self logWithFormat:@"missing component-name of page %@", pageName];
          continue;
        }
        /* ok, add to configurable pages */
        {
          NSDictionary *entry;

          if (label == nil) label = pageName;
          if (icon  == nil) icon  = @"icon_preferences_26x21.gif";

          entry = [NSDictionary dictionaryWithObjectsAndKeys:
                                  pageName,      @"name",
                                  label,         @"labelKey",
                                  bundle,        @"bundle",
                                  componentName, @"componentName",
                                  icon,          @"listicon",
                                  nil];
          [dict setObject:entry forKey:pageName];
        }
      }
      self->availableDockablePages = [dict copyWithZone:[self zone]];
    }
  }
  return self;
}

- (void)dealloc {
  [self->dockedPages            release];
  [self->undockedPages          release];
  [self->availableDockablePages release];;
  [self->dockablePages          release];
  [self->item                   release];
  [self->account                release];
  [self->defaults               release];
  [self->language               release];
  [self->timezone               release];
  [super dealloc];
}

/* notifications */

- (void)awake {
  [super awake];
  self->isRoot = [[self session] activeAccountIsRoot];
}

- (void)sleep {
  [self setItem:nil];
  [super sleep];
}

/* accessors */

- (BOOL)isEditorPage {
  return YES;
}

- (BOOL)isRoot {
  return self->isRoot;
}

- (BOOL)_isEditable:(NSString *)_defName {
  id obj;

  _defName = [@"rootAccess" stringByAppendingString:_defName];
  obj      = [self->defaults objectForKey:_defName];

  return obj ? [obj boolValue] : YES;
}

- (void)setAccount:(id)_account {
  NSZone *z;
  id     tmp;
  
  z = [self zone];

  ASSIGN(self->dockablePages, nil);
  ASSIGN(self->dockedPages,   nil);
  ASSIGN(self->undockedPages, nil);
  ASSIGN(self->language,      nil);
  ASSIGN(self->timezone,      nil);
  ASSIGN(self->account,       _account);

  tmp = self->defaults;
  
  self->defaults = _account
    ? [self runCommand:@"userdefaults::get",
            @"user", _account, nil]
    : [self runCommand:@"userdefaults::get", nil];

  self->defaults = [self->defaults retain];

  ASSIGN(tmp, nil);

  self->dockablePages =
    [[self->defaults arrayForKey:@"SkyDockablePagesOrdering"] copyWithZone:z];
  
  self->defaultButtonsLeft =
    [[self->defaults objectForKey:@"SkyButtonRowDefaultButtonsLeft"]
                     boolValue];
  self->hideInactiveButtons =
    [[self->defaults objectForKey:@"SkyButtonRowHideInactiveButtons"]
                     boolValue];
  self->buttonTextMode =
    [[self->defaults objectForKey:@"SkyButtonTextMode"] boolValue];

  self->maxClipCount =
    [[self->defaults objectForKey:@"SkyMaxFavoritesCount"] intValue];
  self->maxNavLabelLength =
    [[self->defaults objectForKey:@"SkyMaxNavLabelLength"] intValue];

  self->language =
    [[self->defaults stringForKey:@"language"] copyWithZone:[self zone]];
  self->timezone =
    [[self->defaults stringForKey:@"timezone"] copyWithZone:[self zone]];
  self->showAMPMDates =
    [[self->defaults objectForKey:@"scheduler_AMPM_dates"] boolValue];
  
  self->dockTextMode =
    [[self->defaults objectForKey:@"SkyDockTextMode"] boolValue];
  self->dockShowDesktop =
    [[self->defaults objectForKey:@"SkyDockShowDesktop"] boolValue];
  self->dockShowBookmarks =
    [[self->defaults objectForKey:@"SkyDockShowBookmarks"] boolValue];

  self->isDockablePagesEditable =
    [self _isEditable:@"SkyDockablePagesOrdering"];
  self->isDockShowDesktopEditable   = [self _isEditable:@"SkyDockShowDesktop"];
  self->isDockShowBookmarksEditable =
    [self _isEditable:@"SkyDockShowBookmarks"];
  self->isLanguageEditable = [self _isEditable:@"language"];
  self->isTimezoneEditable = [self _isEditable:@"timezone"];

  /* find out docked/undocked pages */
  
  self->dockedPages   = [[NSMutableArray alloc] initWithCapacity:16];
  self->undockedPages = [[NSMutableArray alloc] initWithCapacity:16];
  {
    NSEnumerator *e;
    NSDictionary *entry;
    NSMutableSet *uniquer;

    uniquer = [NSMutableSet setWithCapacity:32];


    e = [self->dockablePages objectEnumerator];
    while ((entry = [e nextObject])) {
      NSDictionary *cfg;

      if ((cfg = [self->availableDockablePages objectForKey:entry])) {
        [self->dockedPages addObject:cfg];
        [uniquer addObject:entry];
      }
    }
    
    e = [self->availableDockablePages objectEnumerator];
    while ((entry = [e nextObject])) {
      NSString *pageName;

      if ((pageName = [entry objectForKey:@"name"]) == nil)
        continue;

      if (![uniquer containsObject:pageName]) {
        [self->undockedPages addObject:entry];
        [uniquer addObject:pageName];
      }
    }
  }
}
- (id)account {
  return self->account;
}

- (NSString *)accountLabel {
  return [[self session] labelForObject:[self account]];
}

- (void)setItem:(id)_item {
  ASSIGN(self->item, _item);
}
- (id)item {
  return self->item;
}

- (NSString *)itemLabel {
  return [[self labels] valueForKey:[self item]];
}

- (NSArray *)timeZoneNames {
  return ([self->defaults arrayForKey:@"LSTimeZones"]);
}

/* access */

- (BOOL)isDockablePagesEditable {
  return self->isRoot || self->isDockablePagesEditable;
}
- (BOOL)isDefaultButtonsLeftEditable {
  return self->isRoot || [self _isEditable:@"SkyButtonRowDefaultButtonsLeft"];
}
- (BOOL)isHideInactiveButtonsEditable {
  return self->isRoot || [self _isEditable:@"SkyButtonRowHideInactiveButtons"];
}
- (BOOL)isButtonTextModeEditable {
  return self->isRoot || [self _isEditable:@"SkyButtonTextMode"];
}

- (BOOL)isMaxClipCountEditable {
  return self->isRoot || [self _isEditable:@"SkyMaxFavoritesCount"];
}

- (BOOL)isDockTextModeEditable {
  return self->isRoot || [self _isEditable:@"SkyDockTextMode"];
}
- (BOOL)isDockShowBookmarksEditable {
  return self->isRoot || self->isDockShowBookmarksEditable;
}
- (BOOL)isDockShowDesktopEditable {
  return self->isRoot || self->isDockShowDesktopEditable;
}
- (BOOL)isLanguageEditable {
  return self->isRoot || self->isLanguageEditable;
}
- (BOOL)isTimezoneEditable {
  return self->isRoot || self->isTimezoneEditable;
}

/* root access */

- (void)setIsDockablePagesEditableRoot:(BOOL)_flag {
  self->isDockablePagesEditable = _flag;
}
- (BOOL)isDockablePagesEditableRoot {
  return self->isDockablePagesEditable;
}

- (void)setIsDockShowBookmarksEditableRoot:(BOOL)_flag {
  self->isDockShowBookmarksEditable = _flag;
}
- (BOOL)isDockShowBookmarksEditableRoot {
  return self->isDockShowBookmarksEditable;
}

- (void)setIsDockShowDesktopEditableRoot:(BOOL)_flag {
  self->isDockShowDesktopEditable = _flag;
}
- (BOOL)isDockShowDesktopEditableRoot {
  return self->isDockShowDesktopEditable;
}

- (void)setIsLanguageEditableRoot:(BOOL)_flag {
  self->isLanguageEditable = _flag;
}
- (BOOL)isLanguageEditableRoot {
  return self->isLanguageEditable;
}

- (void)setIsTimezoneEditableRoot:(BOOL)_flag {
  self->isTimezoneEditable = _flag;
}
- (BOOL)isTimezoneEditableRoot {
  return self->isTimezoneEditable;
}

/* dockable pages */

- (NSString *)dockItemLabel {
  id di;

  if ((di = [self item])) {
    NSString *tmp, *tableName;

    tableName = [[di objectForKey:@"bundle"] bundleName];
    
    if (tableName == nil)
      tableName = [[NGBundle bundleForClass:[self class]] bundleName];
    
    tmp = [di objectForKey:@"labelKey"];
    if (tmp == nil) tmp = [di objectForKey:@"name"];
    if (tmp == nil) tmp = [di objectForKey:@"componentName"];

    if (tmp) {
      di = [[[self application]
                   resourceManager]
                   stringForKey:tmp
                   inTableNamed:tableName
                   withDefaultValue:nil
                   languages:[(WOSession *)[self session] languages]];
      if (di == nil)
        di = tmp;
    }
    else
      di = nil;
  }
  return di;
}

- (id)showDockPage {
  id page;

  page = [[self item] objectForKey:@"componentName"];
  if (page)
    page = [self pageWithName:page];

  if (page) {
    [[[self session] navigation] leavePage];
    [[[self session] navigation] enterPage:page];
  }
  return page;
}

- (NSDictionary *)availableDockablePages {
  return self->availableDockablePages;
}

- (NSArray *)dockablePages {
  return self->dockablePages;
}

- (NSArray *)dockedPages {
  return self->dockedPages;
}
- (NSArray *)undockedPages {
  return self->undockedPages;
}

- (id)addPageAtTop {
  unsigned int idx;
  id           page;

  page = [[[self item] retain] autorelease];
  idx  = [self->undockedPages indexOfObjectIdenticalTo:page];
  
  if (idx == NSNotFound) {
    [self logWithFormat:@"missing dock page %@", page];
    return nil;
  }
  [self->undockedPages removeObjectAtIndex:idx];
  [self->dockedPages   insertObject:page atIndex:0];
  
  return nil;
}
- (id)addPageAtBottom {
  unsigned int idx;
  id           page;

   page = [[[self item] retain] autorelease];
  idx   = [self->undockedPages indexOfObjectIdenticalTo:page];
  
  if (idx == NSNotFound) {
    [self logWithFormat:@"missing dock page %@", page];
    return nil;
  }
  [self->undockedPages removeObjectAtIndex:idx];
  [self->dockedPages   addObject:page];
  return nil;
}

- (id)movePageUp {
  unsigned int idx;
  id           page;

  page = [[[self item] retain] autorelease];
  idx  = [self->dockedPages indexOfObjectIdenticalTo:page];
  
  if (idx == NSNotFound) {
    [self logWithFormat:@"missing dock page %@", page];
    return nil;
  }
  if (idx == 0)
    /* top page */
    return nil;

  [self->dockedPages removeObjectAtIndex:idx];
  [self->dockedPages insertObject:page atIndex:(idx - 1)];
  
  return nil;
}
- (id)movePageDown {
  unsigned int idx, cnt;
  id           page;

  page = [[[self item] retain] autorelease];
  idx  = [self->dockedPages indexOfObjectIdenticalTo:page];
  
  if (idx == NSNotFound) {
    [self logWithFormat:@"missing dock page %@", page];
    return nil;
  }
  cnt = [self->dockedPages count];

  if (idx == cnt - 1) {
    /* last page */
    return nil;
  }
  [self->dockedPages removeObjectAtIndex:idx];
  [self->dockedPages insertObject:page atIndex:(idx + 1)];
  
  return nil;
}
- (id)removeDockPage {
  unsigned int idx;
  id           page;

  page = [[[self item] retain] autorelease];
  idx  = [self->dockedPages indexOfObjectIdenticalTo:page];
  
  if (idx == NSNotFound) {
    [self logWithFormat:@"missing dock page %@", page];
    return nil;
  }
  [self->dockedPages removeObjectAtIndex:idx];
  [self->undockedPages addObject:page];
  return nil;
}

/* values */

- (void)setMaxClipCount:(int)_count {
  if (_count < 1)  _count = 1;
  if (_count > 20) _count = 20;
  self->maxClipCount = _count;
}
- (int)maxClipCount {
  return self->maxClipCount;
}
- (void)setMaxNavLabelLength:(int)_count {
  if (_count < 10)  _count = 10;
  if (_count > 200) _count = 200;
  self->maxNavLabelLength = _count;
}
- (int)maxNavLabelLength {
  return self->maxNavLabelLength;
}

- (void)setDefaultButtonsLeft:(BOOL)_value {
  self->defaultButtonsLeft = _value;
}
- (BOOL)defaultButtonsLeft {
  return self->defaultButtonsLeft;
}

- (void)setHideInactiveButtons:(BOOL)_value {
  self->hideInactiveButtons = _value;
}
- (BOOL)hideInactiveButtons {
  return self->hideInactiveButtons;
}

- (void)setButtonTextMode:(BOOL)_value {
  self->buttonTextMode = _value;
}
- (BOOL)buttonTextMode {
  return self->buttonTextMode;
}

- (void)setDockTextMode:(BOOL)_value {
  self->dockTextMode = _value;
}
- (BOOL)dockTextMode {
  return self->dockTextMode;
}

- (void)setDockShowBookmarks:(BOOL)_value {
  self->dockShowBookmarks = _value;
}
- (BOOL)dockShowBookmarks {
  return self->dockShowBookmarks;
}

- (void)setDockShowDesktop:(BOOL)_value {
  self->dockShowDesktop = _value;
}
- (BOOL)dockShowDesktop {
  return self->dockShowDesktop;
}

- (void)setLanguage:(NSString*)_value {
  ASSIGNCOPY(self->language,_value);
}
- (NSString*)language {
  return self->language;
}

- (BOOL)isTimeZoneEnabled {
  // TODO: deprecated
  return YES;
}

- (void)setTimezone:(NSString*)_value {
  ASSIGNCOPY(self->timezone,_value);  
}
- (NSString*)timezone {
  return self->timezone;
}

- (void)setShowAMPMDates:(BOOL)_flag {
  self->showAMPMDates = _flag;
}
- (BOOL)showAMPMDates {
  return self->showAMPMDates;
}

/* actions */

- (id)cancel {
  [self leavePage];
  return nil;
}

- (id)save {
  static NSNumber *YesNumber = nil;
  static NSNumber *NoNumber   = nil;

  id       uid;
  BOOL     reloadDock;

  if (YesNumber == nil)
    YesNumber = [[NSNumber numberWithBool:YES] retain];

  if (NoNumber == nil)
    NoNumber  = [[NSNumber numberWithBool:NO] retain];

  reloadDock = NO;
  uid        = [[self account] valueForKey:@"companyId"];

  if ([self isDockablePagesEditable]) {
    NSMutableArray *array;
    NSEnumerator   *e;
    NSDictionary   *entry;

    array = [NSMutableArray arrayWithCapacity:16];
    e     = [self->dockedPages objectEnumerator];
    
    while ((entry = [e nextObject])) {
      NSString *pageName;

      pageName = [entry objectForKey:@"name"];

      if (pageName)
        [array addObject:pageName];
    }
    reloadDock = YES;
    
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyDockablePagesOrdering",
            @"value",    array,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isDefaultButtonsLeftEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyButtonRowDefaultButtonsLeft",
            @"value",    self->defaultButtonsLeft ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isHideInactiveButtonsEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyButtonRowHideInactiveButtons",
            @"value",    self->hideInactiveButtons ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isButtonTextModeEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyButtonTextMode",
            @"value",    self->buttonTextMode ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isDockTextModeEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyDockTextMode",
            @"value",    self->dockTextMode ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isDockShowBookmarksEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyDockShowBookmarks",
            @"value",    self->dockShowBookmarks ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isDockShowDesktopEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyDockShowDesktop",
            @"value",    self->dockShowDesktop ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isMaxClipCountEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"SkyMaxFavoritesCount",
            @"value",    [NSNumber numberWithInt:[self maxClipCount]],
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  [self runCommand:@"userdefaults::write",
          @"key",      @"SkyMaxNavLabelLength",
          @"value",    [NSNumber numberWithInt:[self maxNavLabelLength]],
          @"defaults", self->defaults,
          @"userId",   uid,
          nil];
  if ([self isLanguageEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"language",
            @"value",    [self language],
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  if ([self isTimezoneEditable]) {
    [self runCommand:@"userdefaults::write",
            @"key",      @"timezone",
            @"value",    [self timezone],
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  [self runCommand:@"userdefaults::write",
        @"key",   @"scheduler_AMPM_dates",
        @"value", [NSNumber numberWithBool:[self showAMPMDates]],
        @"defaults", self->defaults,
        @"userId", uid,
        nil];
  /* save access stuff */
  if (self->isRoot) {
    [self runCommand:@"userdefaults::write",
            @"key", @"rootAccessSkyDockablePagesOrdering",
            @"value", self->isDockablePagesEditable ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
    [self runCommand:@"userdefaults::write",
            @"key",      @"rootAccessSkyDockShowDesktop",
            @"value",    self->isDockShowDesktopEditable ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
    [self runCommand:@"userdefaults::write",
            @"key",     @"rootAccessSkyDockShowBookmarks",
            @"value",   self->isDockShowBookmarksEditable ? YesNumber : NoNumber,
            @"defaults",self->defaults,
            @"userId",  uid,
            nil];
    [self runCommand:@"userdefaults::write",
            @"key",      @"rootAccesslanguage",
            @"value",    self->isLanguageEditable ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
    [self runCommand:@"userdefaults::write",
            @"key",      @"rootAccesstimezone",
            @"value",    self->isTimezoneEditable ? YesNumber : NoNumber,
            @"defaults", self->defaults,
            @"userId",   uid,
            nil];
  }
  [self postChange:LSWUpdatedAccountNotificationName onObject:[self account]];

  if (reloadDock) {
    [[NSNotificationCenter defaultCenter]
                           postNotificationName:@"SkyDockReload"
                           object:self->defaults];
  }
  return [self leavePage];
}

@end /* SkyDisplayPreferences */
