/*
  Copyright (C) 2000-2003 SKYRIX Software AG

  This file is part of OGo

  OGo is free software; you can redistribute it and/or modify it under
  the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  OGo is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with OGo; see the file COPYING.  If not, write to the
  Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/
// $Id: SkySchedulerSelectPopUp.m,v 1.2 2003/07/11 16:46:01 jan Exp $

#include <OGoFoundation/LSWComponent.h>

/*
  This component generates a PopUp containing groups and accounts. It should
  be extended to allow any kind of person.

    Input parameters:

      showAccount     - should include the login account in the list
      showTeams       - should include the teams in the list
      showSearchField - add ability to search for persons
      fetchGlobalIDs  - work with EOGlobalIDs
      onChange        - JavaScript to execute in the PopUp's onChange attribute

    In/Output parameters:
      
      selectedCompany  - the company which was selected by the user
    
    Output parameters:
      
      selectedGlobalID - the gid of the company which was selected by the user
*/

@interface NSObject(Private)
- (id)commandContext;
@end;

@class NSArray, NSString, NSMutableArray;

@interface SkySchedulerSelectPopUp : LSWComponent
{
  BOOL     fetchGlobalIDs;
  
  /* temporary state */
  NSMutableArray  *possibleItems;
  id              item;
  id              selectedItem;
  id              onChange;
  NSArray         *preSelectedItems;  
  
  BOOL      meToo;
  int       meTooCond;
  NSString *searchText;
  int       maxLabelLength;

  NSArray  *idsForPerson;
  NSArray  *idsForTeams;
  NSArray  *resources;
}

static NSArray *_getGIDSforIds(SkySchedulerSelectPopUp *self, NSArray *_ids,
                               NSString *_entityName);

- (void)_initializePreSelectedItems;
@end


#include <GDLAccess/GDLAccess.h>
#include <OGoFoundation/OGoFoundation.h>
#include <LSFoundation/LSFoundation.h>
#include "common.h"

static int compareItems(id part1, id part2, void *context);

@implementation SkySchedulerSelectPopUp

- (id)init {
  if ((self = [super init])) {
    self->meToo         = YES;
    self->meTooCond     = -1;
    self->possibleItems = nil;
    self->item          = nil;
    self->selectedItem  = nil;
    self->idsForTeams   = nil;
    self->idsForPerson  = nil;
    self->resources     = nil;
    self->maxLabelLength = -1;
  }
  return self;
}

- (void)dealloc {
  RELEASE(self->item);
  RELEASE(self->preSelectedItems);
  RELEASE(self->possibleItems);
  RELEASE(self->idsForTeams);
  RELEASE(self->idsForPerson); 
  RELEASE(self->resources);
  [super dealloc];
}

/* accessors */

- (id)possibleItems {
  return self->possibleItems;
}

- (id)onChange {
  return self->onChange;
}
- (void)setOnChange:(id)_id {
  ASSIGN(self->onChange, _id);
}


- (void)setItem:(id)_item {
  ASSIGN(self->item, _item);
}
- (id)item {
  return self->item;
}

- (BOOL)meToo {
  return self->meToo;
}
- (void)setMeToo:(BOOL)_bool {
  self->meToo = _bool;
}

- (void)setSelectedItem:(id)_item {
  if ([_item isNotNull]) {
    ASSIGN(self->selectedItem, _item);
    
    if (![self->possibleItems containsObject:_item])
      [self->possibleItems addObject:_item];
  }
}
- (id)selectedItem {
  return self->selectedItem;
}

- (BOOL)meTooCond {
  id ac = nil;

  if (self->meTooCond == -1) {
    self->meTooCond = 1;
    ac = [self->session activeAccount];

    if ([self->selectedItem isKindOfClass:[NSString class]] == YES)
      self->meTooCond = 1;

    else if ([self->selectedItem isKindOfClass:[NSDictionary class]] == NO)
      self->meTooCond = 0;
    
    else if ([[[self->selectedItem valueForKey:@"globalID"] entityName]
                                      isEqualToString:@"Team"]) {
      NSEnumerator *enumerator = nil;
      id           obj         = nil;

      enumerator = [[self runCommand:@"account::teams", @"account", ac, nil]
                          objectEnumerator];
      while ((obj = [enumerator nextObject]))
        if ([[obj valueForKey:@"companyId"] isEqual:
                                            [self->selectedItem
                                               valueForKey:@"companyId"]] == YES)
        {
          self->meTooCond = 0;
          break;
        }
    }
    else if ([[[self->selectedItem valueForKey:@"globalID"] entityName]
                                      isEqualToString:@"Person"]) {
      if ([[ac valueForKey:@"companyId"] isEqual:[self->selectedItem
                                                      valueForKey:@"companyId"]]
          == YES)
        self->meTooCond = 0;
    }
    else {
      NSLog(@"WARNING: unknown selected item %@", self->selectedItem);
    }
  }
  return (self->meTooCond == 1) ? YES : NO;
}

- (int)maxLabelLength {
  if (self->maxLabelLength == -1) {
    self->maxLabelLength = [[[self session] userDefaults]
                                   integerForKey:@"scheduler_popup_maxlength"];
    if (self->maxLabelLength < 20) self->maxLabelLength = 20;
  }
  return self->maxLabelLength;
}

- (NSString *)itemLabel {
  NSString *d   = nil;
  id entityName = nil;
  id l          = nil;
  int maxLen    = [self maxLabelLength];

  l = [self labels];

  if ([self->item isKindOfClass:[NSString class]] == YES) {
    if ([self->item isEqualToString:@""]) {
      d = [l valueForKey:@"emptyEntry"];
    }
    else {
      d = [l valueForKey:@"resCategory"];
      if (d == nil) d = @"resCategory";
      d = [NSString stringWithFormat:@" (%@)", d];
      
      if ([self->item hasSuffix:d]) {
        /* is a resource category */
        d = [NSString stringWithFormat:@"%@: %@",
                      [l valueForKey:@"ResourceCategory"],
                      [[self->item componentsSeparatedByString:@" ("]
                                   objectAtIndex:0]];
      }
      else {
        /* is a resource */      
        d = [NSString stringWithFormat:@"%@: %@",
                      [l valueForKey:@"Resource"], self->item];
      }
    }
  }
  else {
    if ([self->item isKindOfClass:[NSDictionary class]] == YES)
      entityName = [[self->item valueForKey:@"globalID"] entityName];
    else
      entityName = [[self->item entity] name];

    if ([entityName isEqualToString:@"Team"] == YES) {
      /* found team */
      d = [NSString stringWithFormat:@"%@: %@",
                    [l valueForKey:@"Team"],
                    [self->item valueForKey:@"description"]];
    }
    else if ([entityName isEqualToString:@"Person"] == YES) {
      NSString *kind = nil;
      NSString *n    = nil;
      NSString *fn   = nil;
      NSString *ln   = nil;
      
      if ([[self->item valueForKey:@"isAccount"] boolValue] == YES)
        kind = @"Account";
      else
        kind = @"Person";

      fn = [self->item valueForKey:@"firstname"];
      ln = [self->item valueForKey:@"name"];

      kind = [l valueForKey:kind];

      if (![ln isNotNull])
        n = [self->item valueForKey:@"login"];
      else {
        if (![fn isNotNull])
          n = ln;
        else
          n = [NSString stringWithFormat:@"%@ %@", fn, ln];
      }
      d = [NSString stringWithFormat:@"%@: %@", kind, n];
    }
  }
  if ([d length] > maxLen) {
    d = [NSString stringWithFormat:@"%@..", [d substringToIndex:maxLen-2]];
  }
  return d;
}

/* processing */

- (NSArray *)distinctCategories:(NSArray *)_items {
  int i, cnt;
  id  obj;
  NSMutableArray *ma;
  NSString *l;

  ma = [[NSMutableArray alloc] init];
  l = [[self labels] valueForKey:@"resCategory"];

  if (l == nil) l = @"resCategory";
  
  i = 0; cnt = [_items count];
  
  while (i < cnt) {
    NSMutableDictionary *rD;
    NSString            *s;

    obj = [_items objectAtIndex:i++];

    s = [NSString stringWithFormat:@"%@ (%@)",
                  [obj valueForKey:@"category"], l];
    rD = [NSMutableDictionary dictionaryWithCapacity:1];
    [rD setObject:s forKey:@"name"];

    if (![ma containsObject:rD])
      [ma addObject:rD];
  }
  return AUTORELEASE(ma);
}

- (void)_initializePreSelectedItems {
  NSMutableArray *array = nil;
  id             tmp    = nil;
  id             defs   = nil;
  BOOL           reload = NO;

  defs = [[self session] userDefaults];
  
  tmp = [defs arrayForKey:@"scheduler_popup_persons"];
  if ((self->idsForPerson == nil) || ([self->idsForPerson isEqual:tmp] == NO))
    reload = YES;

  ASSIGN(self->idsForPerson, tmp);

  tmp = [defs arrayForKey:@"scheduler_popup_teams"];
  if ((self->idsForTeams == nil) || ([self->idsForTeams isEqual:tmp] == NO))
    reload = YES;

  ASSIGN(self->idsForTeams, tmp);

  tmp = [defs arrayForKey:@"scheduler_popup_resourceNames"];
  if ((self->resources == nil) || ([self->resources isEqual:tmp] == NO))
    reload = YES;

  {
    NSEnumerator   *enumerator = nil;
    NSMutableArray *r          = nil;
    NSString       *n          = nil;
    NSString       *s          = nil;

    s = [[self labels] valueForKey:@"resCategory"];
    if (s == nil) s = @"resCategory";
    
    r = [NSMutableArray arrayWithCapacity:8];
    
    enumerator = [tmp objectEnumerator];

    while ((n = [enumerator nextObject])) {
      if ([n hasSuffix:@"(resCategory)"]) {
        n = [[n  componentsSeparatedByString:@" ("] objectAtIndex:0];
        [r addObject:[NSString stringWithFormat:@"%@ (%@)", n, s]];
      }
      else 
        [r addObject:n];
    }
    ASSIGN(self->resources, r);
  }

  if (reload == YES) {
    if (self->preSelectedItems != nil) {
      [self->possibleItems removeObjectsInArray:self->preSelectedItems];
      RELEASE(self->preSelectedItems); self->preSelectedItems = nil;
    }
    
    array = [[NSMutableArray alloc] init];
    
    tmp = _getGIDSforIds(self, self->idsForPerson, @"Person");
    if (tmp != nil)
      [array addObjectsFromArray:tmp];
    
    tmp = _getGIDSforIds(self, self->idsForTeams, @"Team");
    if (tmp != nil)
      [array addObjectsFromArray:tmp];

    if (self->resources != nil)
      [array addObjectsFromArray:self->resources];
  
    { /* remove activeAccount */
      NSEnumerator *enumerator = nil;
      id           obj         = nil;
      id           acid        = nil;

      acid = [[[self session] activeAccount] valueForKey:@"companyId"];

      enumerator = [array objectEnumerator];
      while ((obj = [enumerator nextObject])) {
        if ([obj isKindOfClass:[NSString class]] == NO) {
          if ([[obj valueForKey:@"companyId"] isEqual:acid] == YES)
            break;
        }
      }
      if (obj != nil)
        [array removeObject:obj];
    }
    if ([array count] == 0) {
      id tmp = nil;
      tmp = [self runCommand:@"team::extended-search",
                    @"fetchGlobalIDs", [NSNumber numberWithBool:YES],
                    @"description", @"%%", nil];
      if ([tmp count]) {
        tmp = [[self runCommand:@"team::get-by-globalid",
                       @"gids", tmp,
                       @"groupBy", @"globalID",
                       @"attributes", [NSArray arrayWithObject:@"description"],
                       nil]
                     allValues];
        if (tmp)
        [array addObjectsFromArray:tmp];
      }
    }
    
    [array sortUsingFunction:compareItems context:nil];
    
    self->preSelectedItems = [array copy];
    
    [self->possibleItems addObjectsFromArray:self->preSelectedItems];
    
    if ([self->possibleItems containsObject:self->selectedItem] == NO)
      self->selectedItem = [self->possibleItems objectAtIndex:0];
    
    RELEASE(array); array = nil;
  }
}

static NSArray *_getGIDSforIds(SkySchedulerSelectPopUp *self, NSArray *_ids,
                               NSString *_entityName) {
  NSEnumerator *enumerator = nil;
  id           obj         = nil;
  NSArray      *tmp        = nil;
  EOEntity     *entity     = nil;
  id           *objs       = NULL;
  int          cnt         = 0;
  NSArray      *gids       = nil;
  NSString     *command    = nil;
  NSArray      *args       = nil;

  command = [NSString stringWithFormat:@"%@::get-by-globalid", _entityName];

  if ([_entityName isEqualToString:@"Team"] == YES)
    args = [NSArray arrayWithObjects:@"description", nil];
  else if ([_entityName isEqualToString:@"Person"] == YES)
    args = [NSArray arrayWithObjects:@"name", @"firstname", @"isAccount",
                    @"login", nil];
  else
    NSLog(@"WARNING: unknown entityName %@", _entityName);
  
  if ((_ids != nil) && ([_ids count] > 0)) {
    entity     = [[[(id)[self session] commandContext]
                             valueForKey:LSDatabaseKey] entityNamed:_entityName];
    enumerator = [_ids objectEnumerator];
    objs       = malloc(sizeof(id) * [_ids count]);
    while ((obj = [enumerator nextObject])) {
      objs[cnt++] = [entity globalIDForRow:
                            [NSDictionary dictionaryWithObject:obj
                                          forKey:@"companyId"]];
    }
    gids = [[NSArray alloc] initWithObjects:objs count:cnt];
    tmp  = [[self runCommand:command,
                  @"gids", gids,
                  @"groupBy", @"globalID",
                  @"attributes", args, nil] allValues];
    RELEASE(gids);
    free(objs);
    return tmp;
  }
  else
    return [NSArray array];
}

- (void)syncAwake {
  id ac = [[self session] activeAccount];

  [self _initializePreSelectedItems];

  if (self->possibleItems == nil) {
    self->possibleItems = [[NSMutableArray alloc] init];
  
    [self->possibleItems addObject:ac];
    [self->possibleItems addObject:@""];

    if (self->selectedItem != nil) {
      if (([self->preSelectedItems containsObject:self->selectedItem] == NO) &&
          (self->selectedItem != ac)) {
        [self->possibleItems addObject:self->selectedItem];
        [self->possibleItems addObject:@""];
      }
    }
    else {
      self->selectedItem = ac;
      RETAIN(self->selectedItem);
    }
    [self->possibleItems addObjectsFromArray:self->preSelectedItems];    
  }
  //  [self reconfigure];
  [super syncAwake];
}

- (void)sleep {
  //  RELEASE(self->item);  self->item = nil;
  self->meTooCond = -1;
  [super sleep];
}

/* catch nil-assignments to selected-company */

- (void)takeValuesFromRequest:(WORequest *)_request inContext:(WOContext *)_ctx {
  NSMutableSet    *objs       = nil;

 [super takeValuesFromRequest:_request inContext:_ctx];  
  
  if ([self->searchText length] > 0) {
    int             maxSearch    = 0;
    int             cnt          = 0;
    NSNumber        *yesNum      = nil;
    NSArray         *tmp         = nil;
    NSArray         *personAttr  = nil;
    NSArray         *teamAttr    = nil;
    NSArray         *resAttr     = nil;
    NSArray         *foundedObjs = nil;

    personAttr = [[NSArray alloc] initWithObjects:@"name",
                                  @"firstname", @"isAccount", @"login", nil];
    teamAttr   = [[NSArray alloc] initWithObjects:@"description", nil];
    resAttr    = [[NSArray alloc] initWithObjects:@"name", nil];
    yesNum     = [NSNumber numberWithBool:YES];
    maxSearch  = [[[self session] userDefaults]
                         integerForKey:@"scheduler_additional_popup_entries"];

    /* first fetch accounts */
    objs = [[NSMutableSet alloc] init];

    if (cnt < maxSearch) {
      tmp = [self runCommand:@"team::extended-search",
                  @"fetchGlobalIDs", yesNum,
                  @"operator", @"OR",
                  @"description", self->searchText,
                  @"maxSearchCount", [NSNumber numberWithInt:(maxSearch - cnt)],
                  nil];
      if (tmp != nil) {
        tmp = [[self runCommand:@"team::get-by-globalID",
                     @"gids", tmp,
                     @"groupBy", @"globalID",
                     @"attributes", teamAttr, nil] allValues];
        [objs addObjectsFromArray:tmp];        
      }
      cnt = [objs count];
    }
    if (cnt < maxSearch) {
      tmp = [self runCommand:@"account::extended-search",
                  @"fetchGlobalIDs", yesNum,
                  @"operator", @"OR",
                  @"name", self->searchText,
                  @"firstname",      self->searchText,
                  @"description",    self->searchText,
                  @"login",          self->searchText,
                  @"maxSearchCount", [NSNumber numberWithInt:(maxSearch - cnt)],
                  nil];
    
      if (tmp != nil) {
        tmp = [[self runCommand:@"person::get-by-globalID",
                     @"gids", tmp,
                     @"groupBy", @"globalID",
                     @"attributes", personAttr, nil] allValues];
        [objs addObjectsFromArray:tmp];
        {
          /* remove activeAccount */
          NSEnumerator *enumerator = nil;
          id           obj         = nil;
          id           acId        = nil;

          acId = [[[self session] activeAccount] valueForKey:@"companyId"];

          enumerator = [obj objectEnumerator];

          while ((obj = [enumerator nextObject])) {
            if ([obj isEqual:acId] == YES)
              break;
          }
          if (obj != nil)
            [objs removeObject:obj];
        }
      }
      cnt = [objs count];
    }
    if (cnt < maxSearch) {
      tmp = [self runCommand:@"person::extended-search",
                  @"fetchGlobalIDs", yesNum,
                  @"operator", @"OR",
                  @"name", self->searchText,
                  @"firstname",      self->searchText,
                  @"description",    self->searchText,
                  @"login",          self->searchText,
                  @"withoutAccounts", yesNum,
                  @"maxSearchCount", [NSNumber numberWithInt:(maxSearch - cnt)],
                  nil];
      if (tmp != nil) {
        tmp = [[self runCommand:@"person::get-by-globalID",
                     @"gids", tmp,
                     @"groupBy", @"globalID",
                     @"attributes", personAttr, nil] allValues];
        [objs addObjectsFromArray:tmp];
      }
      cnt = [objs count];
    }
    if (cnt < maxSearch) {
      tmp = [self runCommand:@"appointmentresource::extended-search",
                  @"fetchGlobalIDs",  [NSNumber numberWithBool:YES],
                  @"operator",        @"OR",
                  @"category",        self->searchText,
                  @"maxSearchCount",  [NSNumber numberWithInt:(maxSearch - cnt)],
                  nil];
      if (tmp != nil) {
        tmp = [self runCommand:@"appointmentresource::get-by-globalID",
                    @"gids",         tmp,
                    @"attributes",   [NSArray arrayWithObject:@"category"],
                    nil];
        tmp = [self distinctCategories:tmp];

        if (tmp != nil) {
          [objs addObjectsFromArray:[tmp valueForKey:@"name"]];
        }
      }
      tmp = [self runCommand:@"appointmentresource::extended-search",
                  @"fetchGlobalIDs",  [NSNumber numberWithBool:YES],
                  @"operator",        @"OR",
                  @"name",            self->searchText,
                  @"maxSearchCount",  [NSNumber numberWithInt:(maxSearch - cnt)],
                  nil];
      if (tmp != nil) {
        tmp = [self runCommand:@"appointmentresource::get-by-globalID",
                    @"gids",         tmp,
                    @"attributes",   [NSArray arrayWithObject:@"name"],
                    nil];
      }
      if (tmp != nil) {
        [objs addObjectsFromArray:[tmp valueForKey:@"name"]];
      }
      cnt = [objs count];
    }
    RELEASE(personAttr); personAttr = nil;
    RELEASE(teamAttr);   teamAttr   = nil;
    RELEASE(resAttr);    resAttr    = nil;
    [self->possibleItems removeAllObjects];
    [self->possibleItems addObject:[[self session] activeAccount]];

    if ((objs != nil) && ([objs count] > 0)) {
      foundedObjs = [[objs allObjects] mutableCopy];
      [(NSMutableArray *)foundedObjs sortUsingFunction:compareItems context:nil];
      [self->possibleItems addObject:@""];
      {
        NSEnumerator   *enumerator = nil;
        id             o           = nil;

        enumerator = [foundedObjs objectEnumerator];
        while ((o = [enumerator nextObject])) {
          if ([self->preSelectedItems containsObject:o] == NO) {
            [self->possibleItems addObject:o];
          }
        }
      }
    }
    [self->possibleItems addObject:@""];
    [self->possibleItems addObjectsFromArray:self->preSelectedItems];

    if ([objs count] > 0) {
      RELEASE(self->selectedItem);
      self->selectedItem = [foundedObjs objectAtIndex:0];
      RETAIN(self->selectedItem);
      self->meTooCond = -1;
    }
    else {
      if ([self->possibleItems containsObject:self->selectedItem] == NO) {
        RELEASE(self->selectedItem);
        self->selectedItem = [self->possibleItems objectAtIndex:0];
        RETAIN(self->selectedItem);
        self->meTooCond = -1;
      }
    }
    if (objs != nil) {
      RELEASE(objs); objs = nil;
    }
    RELEASE(self->searchText); self->searchText = nil;
    RELEASE(foundedObjs);      foundedObjs      = nil;
  }
}

- (void)setSearchText:(id)_txt {
  ASSIGN(self->searchText, _txt);
}
- (id)searchText {
  return self->searchText;
}
  
@end /* SkyPersonSelectPopUp */

static Class NSStringClass = Nil;

static NSString *_personName(id _item) {
  NSString *n = nil;
  NSString *f = nil;

  if (NSStringClass == Nil)
    NSStringClass = [NSString class];
  
  n = [_item valueForKey:@"name"];
  f = [_item valueForKey:@"firstname"];

  if (![n isNotNull]) n = @"";
  if (![f isNotNull]) f = @"";
  
  return [NSStringClass stringWithFormat:@"%@ %@", f, n];
}

static int compareItems(id _part1, id _part2, void *_ctx) {
  BOOL     tmp  = NO;
  BOOL     tmp1 = NO;  
  NSString *part1Kind = nil;
  NSString *part2Kind = nil;

  if (NSStringClass == Nil) NSStringClass = [NSString     class];
  
  /* check for resources */
  tmp = [_part2 isKindOfClass:NSStringClass];
  if ([_part1 isKindOfClass:NSStringClass] == YES) {
    if (tmp == YES)
      return [_part1 caseInsensitiveCompare:_part2];
    else
      return NSOrderedDescending;
  }
  if (tmp == YES)
    return NSOrderedAscending;
  
  part1Kind = ([_part1 respondsToSelector:@selector(entity)])
    ? [[_part1 entity] name]
    : [[_part1 valueForKey:@"globalID"] entityName];
  part2Kind = ([_part2 respondsToSelector:@selector(entity)])
    ? [[_part2 entity] name]
    : [[_part2 valueForKey:@"globalID"] entityName];
  
  tmp = [part2Kind isEqualToString:@"Person"];
  
  if ([part1Kind isEqualToString:@"Person"]) {
    if (tmp == YES) {
      tmp1 = [[_part2 valueForKey:@"isAccount"] boolValue];
      
      if ([[_part1 valueForKey:@"isAccount"] boolValue] == YES) {
        if (tmp1 == YES) {
          NSString *pname1, *pname2;

          pname1 = _personName(_part1);
          pname2 = _personName(_part2);
          
          return [pname1 caseInsensitiveCompare:pname2];
        }
        else
          return NSOrderedAscending;
      }
      if (tmp1 == YES) {
        return NSOrderedDescending;
      }
      else { /* person - person */
        NSString *pname1, *pname2;

        pname1 = _personName(_part1);
        pname2 = _personName(_part2);
        
        return [pname1 caseInsensitiveCompare:pname2];
      }
    }
    else {
      return NSOrderedDescending;
    }
  }
  if (tmp == YES) {
    return NSOrderedAscending;
  }
  /* both are teams */
  return [[_part1 valueForKey:@"description"]
                  caseInsensitiveCompare:[_part2 valueForKey:@"description"]];
}
